"""
Test suite for the sync client
"""

from typing import List

import pytest

import biothings_client


@pytest.mark.parametrize(
    "client_name,client_url,class_name",
    [
        (["gene"], "https://mygene.info/v3", "MyGeneInfo"),
        (["variant"], "https://myvariant.info/v1", "MyVariantInfo"),
        (["chem", "drug"], "https://mychem.info/v1", "MyChemInfo"),
        (["disease"], "https://mydisease.info/v1", "MyDiseaseInfo"),
        (["taxon"], "https://t.biothings.io/v1", "MyTaxonInfo"),
        (["geneset"], "https://mygeneset.info/v1", "MyGenesetInfo"),
    ],
)
def test_get_client(client_name: List[str], client_url: str, class_name: str):
    """
    Tests our ability to generate sync clients
    """
    client_name_instances = [biothings_client.get_client(name) for name in client_name]
    client_url_instance = biothings_client.get_client(url=client_url)
    clients = [client_url_instance, *client_name_instances]
    for client in clients:
        assert type(client).__name__ == class_name


@pytest.mark.parametrize(
    "client_name,client_url,class_name",
    [
        ("gene", "https://mygene.info/v3", "MyGeneInfo"),
        ("variant", "https://mychem.info/v1", "MyVariantInfo"),
        ("chem", "https://mychem.info/v1", "MyChemInfo"),
        ("disease", "https://mydisease.info/v1", "MyDiseaseInfo"),
        ("taxon", "https://t.biothings.io/v1", "MyTaxonInfo"),
        ("geneset", "https://mygeneset.info/v1", "MyGenesetInfo"),
    ],
)
def test_generate_settings(client_name: str, client_url: str, class_name: str):
    client_settings = biothings_client.client.base.generate_settings(client_name, url=client_url)
    assert client_settings["class_kwargs"]["_default_url"] == client_url
    assert client_settings["class_name"] == class_name


@pytest.mark.parametrize(
    "client_name",
    (
        "gene",
        "variant",
        "chem",
        "disease",
        "taxon",
        "geneset",
    ),
)
def test_url_protocol(client_name: str):
    """
    Tests that our HTTP protocol modification methods work
    as expected when transforming to either HTTP or HTTPS
    """
    client_instance = biothings_client.get_client(client_name)

    http_protocol = "http://"
    https_protocol = "https://"

    # DEFAULT: HTTPS
    assert client_instance.url.startswith(https_protocol)

    # Transform to HTTP
    client_instance.use_http()
    assert client_instance.url.startswith(http_protocol)

    # Transform back to HTTPS
    client_instance.use_https()
    client_instance.url.startswith(https_protocol)



"""
Custom exceptions for the clients (async and sync)
"""

from typing import List


class OptionalDependencyImportError(ImportError):
    def __init__(self, optional_function_access: str, optional_group: str, libraries: List[str]):
        pip_command = f"`pip install biothings_client[{optional_group}]`"
        message = (
            f"To {optional_function_access} requires the {libraries} library(ies). "
            f"To install run the following command: {pip_command}"
        )
        super().__init__(message)


class CachingNotSupportedError(Exception):
    pass



"""
Tests for exercising the sychronous biothings_client for mychem
"""

import json
import logging
import types

import pytest


import biothings_client
from biothings_client.client.definitions import MyChemInfo
from biothings_client.utils.score import descore


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def test_metadata(chem_client: MyChemInfo):
    meta = chem_client.metadata()
    assert "src" in meta
    assert "stats" in meta
    assert "total" in meta["stats"]


def test_getchem(chem_client: MyChemInfo):
    c = chem_client.getchem("ZRALSGWEFCBTJO-UHFFFAOYSA-N")
    assert c["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"
    assert c["chebi"]["name"] == "guanidine"


def test_getchem_with_fields(chem_client: MyChemInfo):
    c = chem_client.getchem("7AXV542LZ4", fields="chebi.name,chembl.inchi_key,pubchem.cid")
    assert "_id" in c
    assert "chebi" in c
    assert "name" in c["chebi"]
    assert "chembl" in c
    assert "inchi_key" in c["chembl"]
    assert "pubchem" in c
    assert "cid" in c["pubchem"]


def test_curie_id_query(chem_client: MyChemInfo):
    """
    Tests the annotation endpoint support for the biolink CURIE ID.

    If support is enabled then we should retrieve the exact same document for all the provided
    queries
    """

    curie_id_testing_collection = [
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "CHEMBL297569",
            "CHEMBL.COMPOUND:CHEMBL297569",
            "chembl.compound:CHEMBL297569",
            "cHEmbl.ComPOUND:CHEMBL297569",
            "chembl.molecule_chembl_id:CHEMBL297569",
        ),
        (
            "AKUPVPKIFATOBM-UHFFFAOYSA-N",
            "120933777",
            120933777,
            "PUBCHEM.COMPOUND:120933777",
            "pubchem.compound:120933777",
            "PuBcHEm.COMPound:120933777",
            "pubchem.cid:120933777",
        ),
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "CHEBI:CHEBI:57966",
            "chebi:CHEBI:57966",
            "CheBi:CHEBI:57966",
            "chebi.id:CHEBI:57966",
        ),
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "11P2JDE17B",
            "UNII:11P2JDE17B",
            "unii:11P2JDE17B",
            "uNIi:11P2JDE17B",
            "unii.unii:11P2JDE17B",
        ),
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "dB03107",
            "DRUGBANK:dB03107",
            "drugbank:dB03107",
            "DrugBaNK:dB03107",
            "drugbank.id:dB03107",
        ),
    ]

    aggregation_query_groups = []
    for query_collection in curie_id_testing_collection:
        query_result_storage = []
        for similar_query in query_collection:
            query_result = chem_client.getchem(_id=similar_query)
            query_result_storage.append(query_result)

        results_aggregation = [query == query_result_storage[0] for query in query_result_storage[1:]]

        query_result_mapping = dict(zip(query_collection[1:], results_aggregation))
        logger.debug(
            "Comparison to first term %s ->\n%s", query_collection[0], json.dumps(query_result_mapping, indent=4)
        )

        if all(results_aggregation):
            logger.info("Query group %s succeeded", query_collection)
        else:
            logger.error("Query group %s failed", query_collection)

        aggregation_query_groups.append(all(results_aggregation))
    assert all(aggregation_query_groups)


def test_multiple_curie_id_query(chem_client: MyChemInfo):
    """
    Tests the annotations endpoint support for the biolink CURIE ID.

    Batch query testing against the POST endpoint to verify that the CURIE ID can work with
    multiple

    If support is enabled then we should retrieve the exact same document for all the provided
    queries
    """
    curie_id_testing_collection = [
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "CHEMBL297569",
            "CHEMBL.COMPOUND:CHEMBL297569",
            "chembl.compound:CHEMBL297569",
            "cHEmbl.ComPOUND:CHEMBL297569",
            "chembl.molecule_chembl_id:CHEMBL297569",
        ),
        (
            "AKUPVPKIFATOBM-UHFFFAOYSA-N",
            "120933777",
            120933777,
            "PUBCHEM.COMPOUND:120933777",
            "pubchem.compound:120933777",
            "PuBcHEm.COMPound:120933777",
            "pubchem.cid:120933777",
        ),
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "CHEBI:CHEBI:57966",
            "chebi:CHEBI:57966",
            "CheBi:CHEBI:57966",
            "chebi.id:CHEBI:57966",
        ),
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "11P2JDE17B",
            "UNII:11P2JDE17B",
            "unii:11P2JDE17B",
            "uNIi:11P2JDE17B",
            "unii.unii:11P2JDE17B",
        ),
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "dB03107",
            "DRUGBANK:dB03107",
            "drugbank:dB03107",
            "DrugBaNK:dB03107",
            "drugbank.id:dB03107",
        ),
    ]

    results_aggregation = []
    for query_collection in curie_id_testing_collection:
        base_result = chem_client.getchem(_id=query_collection[0])
        query_results = chem_client.getchems(ids=query_collection)
        assert len(query_results) == len(query_collection)

        batch_result = []
        for query_result, query_entry in zip(query_results, query_collection):
            return_query_field = query_result.pop("query")
            assert return_query_field == str(query_entry)
            batch_result.append(base_result == query_result)

        aggregate_result = all(results_aggregation)

        query_result_mapping = dict(zip(query_collection[1:], results_aggregation))
        logger.debug(
            "Comparison to first term %s ->\n%s", query_collection[0], json.dumps(query_result_mapping, indent=4)
        )

        if aggregate_result:
            logger.info("Query group %s succeeded", query_collection)
        else:
            logger.error("Query group %s failed", query_collection)

        results_aggregation.append(aggregate_result)
    assert all(results_aggregation)


@pytest.mark.xfail(reason="pubchem CID not working yet")
def get_getdrug(chem_client: MyChemInfo):
    c = chem_client.getdrug("CHEMBL1308")
    assert c["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"
    c = chem_client.getdrug("7AXV542LZ4")
    assert c["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"
    c = chem_client.getdrug("CHEBI:6431")
    assert c["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"

    # PubChem CID not working yet
    c = chem_client.getdrug("CID:1990")
    assert c["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"
    c = chem_client.getdrug("1990")
    assert c["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


def test_getchems(chem_client: MyChemInfo):
    c_li = chem_client.getchems(
        ["KTUFNOKKBVMGRW-UHFFFAOYSA-N", "HXHWSAZORRCQMX-UHFFFAOYSA-N", "DQMZLTXERSFNPB-UHFFFAOYSA-N"]
    )
    assert len(c_li) == 3
    assert c_li[0]["_id"] == "KTUFNOKKBVMGRW-UHFFFAOYSA-N"
    assert c_li[1]["_id"] == "HXHWSAZORRCQMX-UHFFFAOYSA-N"
    assert c_li[2]["_id"] == "DQMZLTXERSFNPB-UHFFFAOYSA-N"


def test_query(chem_client: MyChemInfo):
    qres = chem_client.query("chebi.name:albendazole", size=5)
    assert "hits" in qres
    assert len(qres["hits"]) == 5


@pytest.mark.xfail(reason="Drugbank was removed")
def test_query_drugbank(chem_client: MyChemInfo):
    qres = chem_client.query("drugbank.id:DB00536")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


def test_query_chebi(chem_client: MyChemInfo):
    qres = chem_client.query(r"chebi.id:CHEBI\:42820")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


def test_query_chembl(chem_client: MyChemInfo):
    qres = chem_client.query('chembl.smiles:"CC(=O)NO"')
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "RRUDCFGSUDOHDG-UHFFFAOYSA-N"


@pytest.mark.xfail(reason="drugcentral query not working yet")
def test_query_drugcentral(chem_client: MyChemInfo):
    qres = chem_client.query("drugcentral.drug_use.contraindication.umls_cui:C0023530", fields="drugcentral", size=50)
    assert "hits" in qres
    assert len(qres["hits"]) == 50

    # not working yet
    qres = chem_client.query("drugcentral.xrefs.kegg_drug:D00220")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


def test_query_pubchem(chem_client: MyChemInfo):
    qres = chem_client.query("pubchem.molecular_formula:C2H5NO2", fields="pubchem", size=20)
    assert "hits" in qres
    assert len(qres["hits"]) == 20

    qres = chem_client.query('pubchem.inchi:"InChI=1S/CH5N3/c2-1(3)4/h(H5,2,3,4)"')
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


def test_query_ginas(chem_client: MyChemInfo):
    qres = chem_client.query("ginas.approvalID:JU58VJ6Y3B")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


def test_query_pharmgkb(chem_client: MyChemInfo):
    qres = chem_client.query("pharmgkb.id:PA164781028")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


def test_query_ndc(chem_client: MyChemInfo):
    qres = chem_client.query('ndc.productndc:"27437-051"')
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "KPQZUUQMTUIKBP-UHFFFAOYSA-N"


@pytest.mark.xfail
def test_query_sider(chem_client: MyChemInfo):
    qres = chem_client.query("sider.meddra.umls_id:C0232487", fields="sider", size=5)
    assert "hits" in qres
    assert len(qres["hits"]) == 5

    # Temp disable this check till we fix the data issue
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


def test_query_unii(chem_client: MyChemInfo):
    qres = chem_client.query("unii.unii:JU58VJ6Y3B")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


def test_query_aeolus(chem_client: MyChemInfo):
    qres = chem_client.query("aeolus.rxcui:50675")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


def test_query_fetch_all(chem_client: MyChemInfo):
    # fetch_all won't work when caching is used.
    q = "drugcentral.drug_use.contraindication.umls_cui:C0023530"
    qres = chem_client.query(q, size=0)
    total = qres["total"]

    qres = chem_client.query(q, fields="drugcentral.drug_use", fetch_all=True)
    assert isinstance(qres, types.GeneratorType)
    assert total == len(list(qres))


def test_querymany(chem_client: MyChemInfo):
    qres = chem_client.querymany(["ZRALSGWEFCBTJO-UHFFFAOYSA-N", "RRUDCFGSUDOHDG-UHFFFAOYSA-N"], verbose=False)
    assert len(qres) == 2

    qres = chem_client.querymany("ZRALSGWEFCBTJO-UHFFFAOYSA-N,RRUDCFGSUDOHDG-UHFFFAOYSA-N", verbose=False)
    assert len(qres) == 2


def test_querymany_with_scopes(chem_client: MyChemInfo):
    qres = chem_client.querymany(["CHEBI:31690", "CHEBI:15365"], scopes="chebi.id", verbose=False)
    assert len(qres) == 2

    qres = chem_client.querymany(
        ["CHEMBL374515", "4RZ82L2GY5"], scopes="chembl.molecule_chembl_id,unii.unii", verbose=False
    )
    assert len(qres) >= 2


def test_querymany_fields(chem_client: MyChemInfo):
    qres1 = chem_client.querymany(
        ["CHEBI:31690", "CHEBI:15365"],
        scopes="chebi.id",
        fields=["chebi.name", "unii.registry_number"],
        verbose=False,
    )
    assert len(qres1) == 2

    qres2 = chem_client.querymany(
        ["CHEBI:31690", "CHEBI:15365"], scopes="chebi.id", fields="chebi.name,unii.registry_number", verbose=False
    )
    assert len(qres2) == 2

    assert descore(qres1) == descore(qres2)


def test_querymany_notfound(chem_client: MyChemInfo):
    qres = chem_client.querymany(["CHEBI:31690", "CHEBI:15365", "NA_TEST"], scopes="chebi.id")
    assert len(qres) == 3
    assert qres[2] == {"query": "NA_TEST", "notfound": True}


@pytest.mark.skipif(not biothings_client._PANDAS, reason="pandas library not installed")
def test_querymany_dataframe(chem_client: MyChemInfo):
    from pandas import DataFrame

    query_list1 = [
        "QCYGXOCMWHSXSU-UHFFFAOYSA-N",
        "ADFOMBKCPIMCOO-BTVCFUMJSA-N",
        "DNUTZBZXLPWRJG-UHFFFAOYSA-N",
        "DROLRDZYPMOKLM-BIVLZKPYSA-N",
        "KPBZROQVTHLCDU-GOSISDBHSA-N",
        "UTUUIUQHGDRVPU-UHFFFAOYSA-K",
        "WZWDUEKBAIXVCC-IGHBBLSQSA-N",
        "IAJIIJBMBCZPSW-BDAKNGLRSA-N",
        "NASIOHFAYPRIAC-JTQLQIEISA-N",
        "VGWIQFDQAFSSKA-UHFFFAOYSA-N",
    ]

    qres = chem_client.querymany(query_list1, scopes="_id", fields="pubchem", as_dataframe=True)
    assert isinstance(qres, DataFrame)
    assert "pubchem.inchi" in qres.columns
    assert set(query_list1) == set(qres.index)


def test_querymany_step(chem_client: MyChemInfo):

    query_list1 = [
        "QCYGXOCMWHSXSU-UHFFFAOYSA-N",
        "ADFOMBKCPIMCOO-BTVCFUMJSA-N",
        "DNUTZBZXLPWRJG-UHFFFAOYSA-N",
        "DROLRDZYPMOKLM-BIVLZKPYSA-N",
        "KPBZROQVTHLCDU-GOSISDBHSA-N",
        "UTUUIUQHGDRVPU-UHFFFAOYSA-K",
        "WZWDUEKBAIXVCC-IGHBBLSQSA-N",
        "IAJIIJBMBCZPSW-BDAKNGLRSA-N",
        "NASIOHFAYPRIAC-JTQLQIEISA-N",
        "VGWIQFDQAFSSKA-UHFFFAOYSA-N",
    ]
    qres1 = chem_client.querymany(query_list1, scopes="_id", fields="pubchem")
    default_step = chem_client.step
    chem_client.step = 3
    qres2 = chem_client.querymany(query_list1, scopes="_id", fields="pubchem")
    chem_client.step = default_step
    qres1 = descore(sorted(qres1, key=lambda doc: doc["_id"]))
    qres2 = descore(sorted(qres2, key=lambda doc: doc["_id"]))
    assert qres1 == qres2


def test_get_fields(chem_client: MyChemInfo):
    fields = chem_client.get_fields()
    assert "chembl.inchi_key" in fields.keys()
    assert "pharmgkb.trade_names" in fields.keys()

    fields = chem_client.get_fields("unii")
    assert "unii.molecular_formula" in fields.keys()



MYVARIANT_TOP_LEVEL_JSONLD_URIS = [
    "http://schema.myvariant.info/datasource/cadd",
    "http://schema.myvariant.info/datasource/clinvar",
    "http://schema.myvariant.info/datasource/dbnsfp",
    "http://schema.myvariant.info/datasource/dbsnp",
    "http://schema.myvariant.info/datasource/docm",
    "http://schema.myvariant.info/datasource/emv",
    "http://schema.myvariant.info/datasource/evs",
    "http://schema.myvariant.info/datasource/gwassnps",
    "http://schema.myvariant.info/datasource/mutdb",
    "http://schema.myvariant.info/datasource/snpeff",
]



""" Gene specific utils. """

from collections import OrderedDict

from biothings_client import get_client


def get_homologs(gene_list, fields="all", species="all"):
    """Return the gene-objects for homologs in species for each gene in the input *gene_list*."""
    gene_client = get_client("gene")
    clean_species = [int(s) for s in species.split(",")] if species != "all" else []
    initial = OrderedDict()
    qset = set()

    for gene in gene_client.getgenes(gene_list, fields="homologene", species="all"):
        if gene["_id"] in initial:
            continue
        initial[gene["_id"]] = {}
        for taxid, geneid in gene.get("homologene", {}).get("genes", []):
            if not clean_species or taxid in clean_species:
                initial[gene["_id"]].setdefault(taxid, []).append(geneid)
                qset.add(geneid)
    gene_dict = {g["_id"]: g for g in gene_client.getgenes(iter(qset), fields=fields, species="all")}

    ret = []
    for geneid, homolog_dict in initial.items():
        _ret = {"gene": geneid, "homologs": []}
        for taxid, glist in homolog_dict.items():
            _d = {"taxid": taxid, "genes": []}
            for g in glist:
                if str(g) in gene_dict:
                    _d["genes"].append(gene_dict[str(g)])
            if len(_d["genes"]) > 0:
                _ret["homologs"].append(_d)
        ret.append(_ret)

    return ret



DOCSTRING = {
    "getchem": """Return the chemical/drug object for the give id.
        This is a wrapper for GET query of "/chem/<chem_id>" service.

        :param _id: a chemical/drug id, supports InchiKey, Drugbank ID, ChEMBL ID, ChEBI ID, PubChem CID and UNII.
                    `More about chemical/drug id <http://docs.mychem.info/en/latest/doc/data.html#id-field>`_.
        :param fields: fields to return, a list or a comma-separated string.
                       If not provided or **fields="all"**, all available fields
                       are returned. See `here <http://docs.mychem.info/en/latest/doc/data.html#available-fields>`__
                       for all available fields.

        :return: a chemical/drug object as a dictionary, or None if _id is not found.

        Example:

        >>> mc.getchem("ZRALSGWEFCBTJO-UHFFFAOYSA-N")
        >>> mc.getchem("DB00553", fields="chebi.name,drugbank.id,pubchem.cid")
        >>> mc.getchem("CHEMBL1308", fields=["chebi.name", "drugbank.id", "pubchem.cid"])
        >>> mc.getchem("7AXV542LZ4", fields="unii")
        >>> mc.getchem("CHEBI:6431", fields="chembl.smiles")

        .. Hint:: The supported field names passed to **fields** parameter can be found from
                  any full chemical/drug object (without **fields**, or **fields="all"**). Note that field name supports dot
                  notation for nested data structure as well, e.g. you can pass "drugbank.id" or
                  "chembl.smiles".
        """,
    "getchems": """Return the list of chemical/drug annotation objects for the given list of chemical/drug ids.
        This is a wrapper for POST query of "/chem" service.

        :param ids: a list/tuple/iterable or a string of comma-separated chem/drug ids.
                    `More about chem/drug id <http://docs.mychem.info/en/latest/doc/data.html#id-field>`_.
        :param fields: fields to return, a list or a comma-separated string.
                       If not provided or **fields="all"**, all available fields
                       are returned. See `here <http://docs.mychem.info/en/latest/doc/data.html#available-fields>`__
                       for all available fields.
        :param as_generator:  if True, will yield the results in a generator.
        :param as_dataframe: if True or 1 or 2, return object as DataFrame (requires Pandas).
                                  True or 1: using json_normalize
                                  2        : using DataFrame.from_dict
                                  otherwise: return original json
        :param df_index: if True (default), index returned DataFrame by 'query',
                         otherwise, index by number. Only applicable if as_dataframe=True.

        :return: a list of variant objects or a pandas DataFrame object (when **as_dataframe** is True)

        :ref: http://docs.mychem.info/en/latest/doc/chem_annotation_service.html.

        Example:

        >>> chems = [
        ...     "KTUFNOKKBVMGRW-UHFFFAOYSA-N",
        ...     "HXHWSAZORRCQMX-UHFFFAOYSA-N",
        ...     "DQMZLTXERSFNPB-UHFFFAOYSA-N"
        ... ]
        >>> mc.getchems(chems, fields="pubchem")
        >>> mc.getchems('KTUFNOKKBVMGRW-UHFFFAOYSA-N,DB00553', fields="all")
        >>> mc.getchems(chems, fields='chembl', as_dataframe=True)

        .. Hint:: A large list of more than 1000 input ids will be sent to the backend
                  web service in batches (1000 at a time), and then the results will be
                  concatenated together. So, from the user-end, it's exactly the same as
                  passing a shorter list. You don't need to worry about saturating our
                  backend servers.

        .. Hint:: If you need to pass a very large list of input ids, you can pass a generator
                  instead of a full list, which is more memory efficient.
        """,
    "query": """Return  the query result.
        This is a wrapper for GET query of "/query?q=<query>" service.

        :param q: a query string, detailed query syntax `here <http://docs.mychem.info/en/latest/doc/chem_query_service.html#query-syntax>`__.
        :param fields: fields to return, a list or a comma-separated string.
                       If not provided or **fields="all"**, all available fields
                       are returned. See `here <http://docs.mychem.info/en/latest/doc/data.html#available-fields>`__
                       for all available fields.
        :param size:   the maximum number of results to return (with a cap
                       of 1000 at the moment). Default: 10.
        :param skip:   the number of results to skip. Default: 0.
        :param sort:   Prefix with "-" for descending order, otherwise in ascending order.
                       Default: sort by matching scores in decending order.
        :param as_dataframe: if True or 1 or 2, return object as DataFrame (requires Pandas).
                                  True or 1: using json_normalize
                                  2        : using DataFrame.from_dict
                                  otherwise: return original json
        :param fetch_all: if True, return a generator to all query results (unsorted).  This can provide a very fast
                          return of all hits from a large query.
                          Server requests are done in blocks of 1000 and yielded individually.  Each 1000 block of
                          results must be yielded within 1 minute, otherwise the request will expire at server side.

        :return: a dictionary with returned variant hits or a pandas DataFrame object (when **as_dataframe** is True)
                 or a generator of all hits (when **fetch_all** is True)
        :ref: http://docs.mychem.info/en/latest/doc/chem_query_service.html.

        Example:

        >>> mc.query('drugbank.name:monobenzone')
        >>> mc.query('drugbank.targets.uniprot:P07998')
        >>> mc.query('drugbank.targets.uniprot:P07998 AND _exists_:unii')
        >>> mc.query('chebi.mass:[300 TO 500]')
        >>> mc.query('sider.side_effect.name:anaemia', size=5)
        >>> mc.query('sider.side_effect.name:anaemia', fetch_all=True)

        .. Hint:: By default, **query** method returns the first 10 hits if the matched hits are >10. If the total number
                  of hits are less than 1000, you can increase the value for **size** parameter. For a query returns
                  more than 1000 hits, you can pass "fetch_all=True" to return a `generator <http://www.learnpython.org/en/Generators>`_
                  of all matching hits (internally, those hits are requested from the server-side in blocks of 1000).
        """,
    "querymany": """Return the batch query result.
        This is a wrapper for POST query of "/query" service.

        :param qterms: a list/tuple/iterable of query terms, or a string of comma-separated query terms.
        :param scopes: specify the type (or types) of identifiers passed to **qterms**, either a list or a comma-separated fields to specify type of
                       input qterms, e.g. "dbsnp.rsid", "clinvar.rcv_accession", ["dbsnp.rsid", "cosmic.cosmic_id"].
                       See `here <http://docs.myvariant.info/en/latest/doc/data.html#available-fields>`__ for full list
                       of supported fields.
        :param fields: fields to return, a list or a comma-separated string.
                       If not provided or **fields="all"**, all available fields
                       are returned. See `here <http://docs.myvariant.info/en/latest/doc/data.html#available-fields>`__
                       for all available fields.
        :param returnall:   if True, return a dict of all related data, including dup. and missing qterms
        :param verbose:     if True (default), print out information about dup and missing qterms
        :param as_dataframe: if True or 1 or 2, return object as DataFrame (requires Pandas).
                                  True or 1: using json_normalize
                                  2        : using DataFrame.from_dict
                                  otherwise: return original json
        :param df_index: if True (default), index returned DataFrame by 'query',
                         otherwise, index by number. Only applicable if as_dataframe=True.
        :return: a list of matching variant objects or a pandas DataFrame object.
        :ref: http://docs.myvariant.info/en/latest/doc/variant_query_service.html for available
              fields, extra *kwargs* and more.

        Example:

        >>> mc.querymany(["ZRALSGWEFCBTJO-UHFFFAOYSA-N", "RRUDCFGSUDOHDG-UHFFFAOYSA-N"])
        >>> mc.querymany(["DB00536", 'DB00533'], scopes='drugbank.id')
        >>> mc.querymany(["CHEBI:95222", 'CHEBI:45924', 'CHEBI:33325'], scopes='chebi.id')
        >>> mc.querymany(["CHEMBL1555813", 'CHEMBL22', 'CHEMBL842'], scopes='chembl.molecule_chembl_id')
        >>> mc.querymany(["DB00536", '4RZ82L2GY5'], scopes='drugbank.id,unii.unii')
        >>> mc.querymany(["DB00536", '4RZ82L2GY5'], scopes=['drugbank.id', 'unii.unii'])
        >>> mc.querymany(["DB00536", '4RZ82L2GY5'], scopes=['drugbank.id', 'unii.unii'], fields='drugbank,unii')
        >>> mc.querymany(["DB00536", '4RZ82L2GY5'], scopes=['drugbank.id', 'unii.unii'],
        ...              fields='drugbank.name,unii',as_dataframe=True)

        .. Hint:: :py:meth:`querymany` is perfect for query variants based different ids, e.g. rsid, clinvar ids, etc.

        .. Hint:: Just like :py:meth:`getvariants`, passing a large list of ids (>1000) to :py:meth:`querymany` is perfectly fine.

        .. Hint:: If you need to pass a very large list of input qterms, you can pass a generator
                  instead of a full list, which is more memory efficient.

        """,
    "metadata": """Return a dictionary of MyChem.info metadata, a wrapper for http://mychem.info/v1/metadata

        Example:

        >>> metadata = mv.metadata()

        """,
    "get_fields": """Wrapper for http://mychem.info/v1/metadata/fields

        :param search_term: a case insensitive string to search for in available field names.
                            If not provided, all available fields will be returned.


        Example:

        >>> mc.get_fields()
        >>> mc.get_fields("pubchem")
        >>> mc.get_fields("drugbank.targets")

        .. Hint:: This is useful to find out the field names you need to pass to **fields** parameter of other methods.

        """,
}



def get_dotfield(d, df):
    s = set()

    def _helper(_d, _k):
        _f = _k.split(".")
        if len(_f) == 0:
            return
        elif len(_f) == 1:
            _tk = _f[0]
            _nk = ""
        else:
            _tk = _f[0]
            _nk = ".".join(_f[1:])
        if isinstance(_d, list):
            for x in _d:
                _helper(x, _tk)
        elif isinstance(_d, dict):
            _v = _d.get(_tk, None)
            if _v:
                _helper(_v, _nk)
        else:
            s.add(_d)

    _helper(d, df)
    return list(s)


def unordered_chunk_iterator(client, query, join_field, chunk_size=100, query_kwargs=None):
    chunk = []
    join_val_dict = {}
    query_kwargs = query_kwargs or {}
    if query_kwargs.get("fields", None) and query_kwargs["fields"] != "all" and join_field != "_id":
        query_kwargs["fields"] = query_kwargs["fields"].rstrip(", ") + "," + join_field
    for doc in client.query(query, fetch_all=True, **query_kwargs):
        for doc_join_val in get_dotfield(doc, join_field):
            join_val_dict.setdefault(str(doc_join_val).lower(), []).append(len(chunk))
        chunk.append(doc)
        if len(join_val_dict) == chunk_size:
            yield chunk, join_val_dict
            chunk = []
            join_val_dict = {}
    if chunk:
        yield chunk, join_val_dict


def join(
    e1_client,
    e2_client,
    size=10,
    e1_query="__all__",
    e2_query="__all__",
    e1_join_field="_id",
    e2_join_field="_id",
    e1_kwargs=None,
    e2_kwargs=None,
):
    """implements a join with e1 being the outer loop and e2 being the inner loop"""
    ret_chunk = []
    e1_kwargs = e1_kwargs or {}
    e2_kwargs = e2_kwargs or {}
    for outer_doc_chunk, outer_join_val_dict in unordered_chunk_iterator(e1_client, e1_query, e1_join_field, e1_kwargs):
        if outer_doc_chunk:
            inner_query_string = " OR ".join(["{}:{}".format(e2_join_field, x) for x in outer_join_val_dict.keys()])
            if e2_query != "__all__":
                inner_query_string = "((" + inner_query_string + ") AND (" + e2_query + "))"
            if e2_kwargs.get("fields", None) and e2_kwargs["fields"] != "all" and e2_join_field != "_id":
                e2_kwargs["fields"] = e2_kwargs["fields"].rstrip(", ") + "," + e2_join_field
            e2_val_join_dict = {}
            for inner_doc in e2_client.query(inner_query_string, fetch_all=True, **e2_kwargs):
                for doc_join_val in get_dotfield(inner_doc, e2_join_field):
                    e2_val_join_dict.setdefault(str(doc_join_val).lower(), []).append(inner_doc)
            # merge the docs for this chunk
            chunk_intersection = set(list(e2_val_join_dict.keys())).intersection(set(list(outer_join_val_dict.keys())))
            if len(chunk_intersection) > 0:
                for merge_join_field in set(list(e2_val_join_dict.keys())).intersection(
                    set(list(outer_join_val_dict.keys()))
                ):
                    for index in outer_join_val_dict[merge_join_field]:
                        outer_doc_chunk[index].setdefault(e2_client._entity, []).extend(
                            e2_val_join_dict[merge_join_field]
                        )
                for p in {p for v in chunk_intersection for p in outer_join_val_dict[v]}:
                    ret_chunk.append(outer_doc_chunk[p])
                    if len(ret_chunk) == size:
                        yield ret_chunk
                        ret_chunk = []
    if ret_chunk:
        yield ret_chunk



class MyGeneClientMixin:
    def findgenes(self, id_li, **kwargs):
        """
        .. deprecated:: 2.0.0
           Use :py:meth:`querymany` instead. It's kept here as an alias of :py:meth:`querymany` method.

        """
        import warnings

        warnings.warn(
            'Deprecated! Currently an alias of "querymany" method. Use "querymany" method directly.', DeprecationWarning
        )
        return self.querymany(id_li, **kwargs)



import types


def copy_func(f: types.FunctionType, name: str = None):
    """
    return a function with same code, globals, defaults, closure, and
    name (or provide a new name)
    """
    fn = types.FunctionType(f.__code__, f.__globals__, name or f.__name__, f.__defaults__, f.__closure__)
    # in case f was given attrs (note this dict is a shallow copy):
    fn.__dict__.update(f.__dict__)
    return fn



"""
Tests the client caching functionality
"""

import datetime
import logging
from typing import Callable

import pytest

import biothings_client
from biothings_client.client.asynchronous import get_async_client
from biothings_client.client.base import get_client


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@pytest.mark.skipif(not biothings_client._CACHING, reason="caching libraries not installed")
@pytest.mark.parametrize(
    "method,client,function,function_arguments",
    [
        ("GET", "gene", "getgene", {"_id": "1017", "return_raw": True}),
        ("POST", "gene", "getgenes", {"ids": ["1017", "1018"], "return_raw": True}),
        ("GET", "gene", "query", {"q": "cdk2", "return_raw": True}),
        ("POST", "gene", "querymany", {"qterms": ["1017", "695"], "return_raw": True}),
        (
            "GET",
            "chem",
            "getdrug",
            {"_id": "ZRALSGWEFCBTJO-UHFFFAOYSA-N", "return_raw": True},
        ),
        (
            "POST",
            "chem",
            "getdrugs",
            {"ids": ["ZRALSGWEFCBTJO-UHFFFAOYSA-N", "RRUDCFGSUDOHDG-UHFFFAOYSA-N"], "return_raw": True},
        ),
        (
            "GET",
            "chem",
            "query",
            {"q": "chebi.name:albendazole", "size": 5, "return_raw": True},
        ),
        (
            "POST",
            "chem",
            "querymany",
            {"qterms": ["CHEBI:31690", "CHEBI:15365"], "scopes": "chebi.id", "return_raw": True},
        ),
        ("GET", "variant", "getvariant", {"_id": "chr9:g.107620835G>A", "return_raw": True}),
        ("POST", "variant", "getvariants", {"ids": ["chr9:g.107620835G>A", "chr1:g.876664G>A"], "return_raw": True}),
        ("GET", "variant", "query", {"q": "dbsnp.rsid:rs58991260", "return_raw": True}),
        (
            "POST",
            "variant",
            "querymany",
            {"qterms": ["rs58991260", "rs2500"], "scopes": "dbsnp.rsid", "return_raw": True},
        ),
        ("GET", "geneset", "getgeneset", {"_id": "WP100", "return_raw": True}),
        ("POST", "geneset", "getgenesets", {"ids": ["WP100", "WP101"], "return_raw": True}),
        ("GET", "geneset", "query", {"q": "wnt", "fields": "name,count,source,taxid", "return_raw": True}),
        (
            "POST",
            "geneset",
            "querymany",
            {"qterms": ["wnt", "jak-stat"], "fields": "name,count,source,taxid", "return_raw": True},
        ),
    ],
)
def test_sync_caching(method: str, client: str, function: str, function_arguments: dict):
    """
    Tests sync caching methods for a variety of data
    """
    try:
        client_instance = get_client(client)
        client_instance._build_http_client()
        client_instance.set_caching()
        assert client_instance.caching_enabled
        client_instance.clear_cache()

        client_callback = getattr(client_instance, function)
        assert isinstance(client_callback, Callable)
        cold_response = client_callback(**function_arguments)
        hot_response = client_callback(**function_arguments)

        client_instance.stop_caching()
        forced_cold_response = client_callback(**function_arguments)
        client_instance.set_caching()
        cold_response2 = client_callback(**function_arguments)
        client_instance.clear_cache()
        forced_cold_response2 = client_callback(**function_arguments)
        hot_response2 = client_callback(**function_arguments)

        cold_responses = [cold_response, cold_response2, forced_cold_response, forced_cold_response2]
        hot_responses = [hot_response, hot_response2]
        for cold_response in cold_responses:
            assert cold_response.status_code == 200
            assert not cold_response.extensions.get("from_cache", False)
            assert not cold_response.extensions.get("revalidated", False)

        for hot_response in hot_responses:
            assert hot_response.status_code == 200
            assert hot_response.extensions["from_cache"]
            assert not hot_response.extensions["revalidated"]
            assert isinstance(hot_response.extensions["cache_metadata"]["created_at"], datetime.datetime)
            assert hot_response.extensions["cache_metadata"]["number_of_uses"] >= 1
            assert all(hot_response.elapsed < cold_response.elapsed for cold_response in cold_responses)

    except Exception as gen_exc:
        client_instance.clear_cache()
        logger.exception(gen_exc)
        raise gen_exc


@pytest.mark.asyncio
@pytest.mark.skipif(not biothings_client._CACHING, reason="caching libraries not installed")
@pytest.mark.parametrize(
    "method,client,function,function_arguments",
    [
        ("GET", "gene", "getgene", {"_id": "1017", "return_raw": True}),
        ("POST", "gene", "getgenes", {"ids": ["1017", "1018"], "return_raw": True}),
        ("GET", "gene", "query", {"q": "cdk2", "return_raw": True}),
        ("POST", "gene", "querymany", {"qterms": ["1017", "695"], "return_raw": True}),
        (
            "GET",
            "chem",
            "getdrug",
            {"_id": "ZRALSGWEFCBTJO-UHFFFAOYSA-N", "return_raw": True},
        ),
        (
            "POST",
            "chem",
            "getdrugs",
            {"ids": ["ZRALSGWEFCBTJO-UHFFFAOYSA-N", "RRUDCFGSUDOHDG-UHFFFAOYSA-N"], "return_raw": True},
        ),
        (
            "GET",
            "chem",
            "query",
            {"q": "chebi.name:albendazole", "size": 5, "return_raw": True},
        ),
        (
            "POST",
            "chem",
            "querymany",
            {"qterms": ["CHEBI:31690", "CHEBI:15365"], "scopes": "chebi.id", "return_raw": True},
        ),
        ("GET", "variant", "getvariant", {"_id": "chr9:g.107620835G>A", "return_raw": True}),
        ("POST", "variant", "getvariants", {"ids": ["chr9:g.107620835G>A", "chr1:g.876664G>A"], "return_raw": True}),
        ("GET", "variant", "query", {"q": "dbsnp.rsid:rs58991260", "return_raw": True}),
        (
            "POST",
            "variant",
            "querymany",
            {"qterms": ["rs58991260", "rs2500"], "scopes": "dbsnp.rsid", "return_raw": True},
        ),
        ("GET", "geneset", "getgeneset", {"_id": "WP100", "return_raw": True}),
        ("POST", "geneset", "getgenesets", {"ids": ["WP100", "WP101"], "return_raw": True}),
        ("GET", "geneset", "query", {"q": "wnt", "fields": "name,count,source,taxid", "return_raw": True}),
        (
            "POST",
            "geneset",
            "querymany",
            {"qterms": ["wnt", "jak-stat"], "fields": "name,count,source,taxid", "return_raw": True},
        ),
    ],
)
async def test_async_caching(method: str, client: str, function: str, function_arguments: dict):
    """
    Tests async caching methods for a variety of data
    """
    try:
        client_instance = get_async_client(client)
        await client_instance._build_http_client()
        await client_instance.set_caching()
        assert client_instance.caching_enabled
        await client_instance.clear_cache()

        client_callback = getattr(client_instance, function)
        assert isinstance(client_callback, Callable)
        cold_response = await client_callback(**function_arguments)
        hot_response = await client_callback(**function_arguments)

        await client_instance.stop_caching()
        forced_cold_response = await client_callback(**function_arguments)
        await client_instance.set_caching()
        cold_response2 = await client_callback(**function_arguments)
        await client_instance.clear_cache()
        forced_cold_response2 = await client_callback(**function_arguments)
        hot_response2 = await client_callback(**function_arguments)

        cold_responses = [cold_response, cold_response2, forced_cold_response, forced_cold_response2]
        hot_responses = [hot_response, hot_response2]
        for cold_response in cold_responses:
            assert cold_response.status_code == 200
            assert not cold_response.extensions.get("from_cache", False)
            assert not cold_response.extensions.get("revalidated", False)

        for hot_response in hot_responses:
            assert hot_response.status_code == 200
            assert hot_response.extensions["from_cache"]
            assert not hot_response.extensions["revalidated"]
            assert isinstance(hot_response.extensions["cache_metadata"]["created_at"], datetime.datetime)
            assert hot_response.extensions["cache_metadata"]["number_of_uses"] >= 1
            assert all(hot_response.elapsed < cold_response.elapsed for cold_response in cold_responses)

    except Exception as gen_exc:
        await client_instance.clear_cache()
        logger.exception(gen_exc)
        raise gen_exc



"""
Mirror of the geneset tests but two main differences:
> asynchronous
> implemented in pytest for asyncio marker
"""

import logging
import types

import pytest


from biothings_client.client.definitions import AsyncMyGenesetInfo
from biothings_client.utils.score import descore


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@pytest.mark.asyncio
async def test_metadata(async_geneset_client: AsyncMyGenesetInfo):
    meta = await async_geneset_client.metadata()
    assert "src" in meta
    assert "stats" in meta
    assert "total" in meta["stats"]


@pytest.mark.asyncio
async def test_getgeneset(async_geneset_client: AsyncMyGenesetInfo):
    gs = await async_geneset_client.getgeneset("WP100")
    assert gs["_id"] == "WP100"
    assert gs["name"] == "Glutathione metabolism"
    assert gs["source"] == "wikipathways"
    assert gs["taxid"] == "9606"
    assert len(gs["genes"]) >= 19
    assert gs["count"] == len(gs["genes"])

    assert "wikipathways" in gs
    assert gs["wikipathways"]["id"] == "WP100"
    assert gs["wikipathways"]["pathway_name"] == "Glutathione metabolism"
    assert gs["wikipathways"]["url"] == "https://www.wikipathways.org/instance/WP100"
    assert gs["wikipathways"]["_license"] == "https://www.wikipathways.org/terms.html"

    assert any((gene.get("mygene_id") == "2937" and gene.get("symbol") == "GSS") for gene in gs["genes"])


@pytest.mark.asyncio
async def test_query_fetch_all(async_geneset_client: AsyncMyGenesetInfo):
    # pdb-->reactome
    # q=source:reactome
    # _exists_:pdb ---> source:reactome

    # fetch_all won't work when caching is used.
    # await async_geneset_client.stop_caching()
    qres = await async_geneset_client.query("source:reactome")
    total = qres["total"]

    qres_generator = await async_geneset_client.query("source:reactome", fields="source,count,name", fetch_all=True)
    assert isinstance(qres_generator, types.AsyncGeneratorType)

    async_count = 0
    async for async_res in qres_generator:
        async_count += 1
    assert total == async_count


@pytest.mark.asyncio
async def test_query_with_fields_as_list(async_geneset_client: AsyncMyGenesetInfo):
    qres1 = await async_geneset_client.query("genes.ncbigene:1017", fields="name,source,taxid")
    qres2 = await async_geneset_client.query("genes.ncbigene:1017", fields=["name", "source", "taxid"])
    assert "hits" in qres1
    assert len(qres1["hits"]) == 10
    assert descore(qres1["hits"]) == descore(qres2["hits"])


@pytest.mark.asyncio
async def test_getgeneset_with_fields(async_geneset_client: AsyncMyGenesetInfo):
    gs = await async_geneset_client.getgeneset("WP100", fields="name,source,taxid,genes.mygene_id,genes.symbol")

    assert "_id" in gs
    assert "name" in gs
    assert "source" in gs
    assert "taxid" in gs

    assert any((gene.get("mygene_id") == "2937" and gene.get("symbol") == "GSS") for gene in gs["genes"])
    assert not any(gene.get("name") for gene in gs["genes"])


@pytest.mark.asyncio
async def test_getgenesets(async_geneset_client: AsyncMyGenesetInfo):
    gs_li = await async_geneset_client.getgenesets(["WP100", "WP101", "WP103"])

    assert len(gs_li) == 3
    assert gs_li[0]["_id"] == "WP100"
    assert gs_li[1]["_id"] == "WP101"
    assert gs_li[2]["_id"] == "WP103"


@pytest.mark.asyncio
async def test_query(async_geneset_client: AsyncMyGenesetInfo):
    qres = await async_geneset_client.query("genes.mygene_id:2937", size=5)
    assert "hits" in qres
    assert len(qres["hits"]) == 5


@pytest.mark.asyncio
async def test_query_default_fields(async_geneset_client: AsyncMyGenesetInfo):
    await async_geneset_client.query(q="glucose")


@pytest.mark.asyncio
async def test_query_field(async_geneset_client: AsyncMyGenesetInfo):
    await async_geneset_client.query(q="genes.ncbigene:1017")


@pytest.mark.asyncio
async def test_species_filter_plus_query(async_geneset_client: AsyncMyGenesetInfo):
    dog = await async_geneset_client.query(q="glucose", species="9615")
    assert dog["hits"][0]["taxid"] == "9615"


@pytest.mark.asyncio
async def test_query_by_id(async_geneset_client: AsyncMyGenesetInfo):
    query = await async_geneset_client.query(q="_id:WP100")
    assert query["hits"][0]["_id"] == "WP100"


@pytest.mark.asyncio
async def test_query_by_name(async_geneset_client: AsyncMyGenesetInfo):
    kinase = await async_geneset_client.query(q="name:kinase")
    assert "kinase" in kinase["hits"][0]["name"].lower()


@pytest.mark.asyncio
async def test_query_by_description(async_geneset_client: AsyncMyGenesetInfo):
    desc = await async_geneset_client.query(q="description:cytosine deamination")
    assert "cytosine" in desc["hits"][0]["description"].lower()
    assert "deamination" in desc["hits"][0]["description"].lower()


@pytest.mark.asyncio
async def test_query_by_source_go(async_geneset_client: AsyncMyGenesetInfo):
    go = await async_geneset_client.query(q="source:go", fields="all")
    assert "go" in go["hits"][0].keys()
    assert go["hits"][0]["source"] == "go"


@pytest.mark.asyncio
async def test_query_by_source_ctd(async_geneset_client: AsyncMyGenesetInfo):
    ctd = await async_geneset_client.query(q="source:ctd", fields="all")
    assert "ctd" in ctd["hits"][0].keys()
    assert ctd["hits"][0]["source"] == "ctd"


@pytest.mark.asyncio
async def test_query_by_source_msigdb(async_geneset_client: AsyncMyGenesetInfo):
    msigdb = await async_geneset_client.query(q="source:msigdb", fields="all")
    assert "msigdb" in msigdb["hits"][0].keys()
    assert msigdb["hits"][0]["source"] == "msigdb"


@pytest.mark.skip(reason="We removed kegg data source for now")
@pytest.mark.asyncio
async def test_query_by_source_kegg(async_geneset_client: AsyncMyGenesetInfo):
    kegg = await async_geneset_client.query(q="source:kegg", fields="all")
    assert "kegg" in kegg["hits"][0].keys()
    assert kegg["hits"][0]["source"] == "kegg"


@pytest.mark.asyncio
async def test_query_by_source_do(async_geneset_client: AsyncMyGenesetInfo):
    do = await async_geneset_client.query(q="source:do", fields="all")
    assert "do" in do["hits"][0].keys()
    assert do["hits"][0]["source"] == "do"


@pytest.mark.asyncio
async def test_query_by_source_reactome(async_geneset_client: AsyncMyGenesetInfo):
    reactome = await async_geneset_client.query(q="source:reactome", fields="all")
    assert "reactome" in reactome["hits"][0].keys()
    assert reactome["hits"][0]["source"] == "reactome"


@pytest.mark.asyncio
async def test_query_by_source_smpdb(async_geneset_client: AsyncMyGenesetInfo):
    smpdb = await async_geneset_client.query(q="source:smpdb", fields="all")
    assert "smpdb" in smpdb["hits"][0].keys()
    assert smpdb["hits"][0]["source"] == "smpdb"



from collections import Counter
from itertools import islice
from typing import Iterable
import logging


logger = logging.getLogger("biothings.client")
logger.setLevel(logging.INFO)


def safe_str(s, encoding="utf-8"):
    """Perform proper encoding if input is an unicode string."""
    try:
        _s = str(s)
    except UnicodeEncodeError:
        _s = s.encode(encoding)
    return _s


def list_itemcnt(li):
    """
    Return number of occurrence for each item in the list.
    """
    return list(Counter(li).items())


def iter_n(iterable, n, with_cnt=False):
    """
    Iterate an iterator by chunks (of n)
    if with_cnt is True, return (chunk, cnt) each time
    """
    it = iter(iterable)
    if with_cnt:
        cnt = 0
    while True:
        chunk = tuple(islice(it, n))
        if not chunk:
            return
        if with_cnt:
            cnt += len(chunk)
            yield (chunk, cnt)
        else:
            yield chunk


def concatenate_list(sequence: Iterable, sep: str = ",", quoted: bool = True) -> str:
    """
    Ingests an iterable sequence and combines all elements into a string

    :param sequence: Iterable sequence for element concatenation
    :param sep: Delimiter to apply to the elements when concatenating, defaults to ","
    :param quoted: Flag to enable quoting of the elements within the string, defaults to True

    :return: Returns a concatenated string of the elements from the input sequence
    :rtype: string
    """
    if isinstance(sequence, (list, tuple)):
        if quoted:
            string_transform = sep.join(['"{}"'.format(safe_str(x)) for x in sequence])
        else:
            string_transform = sep.join(["{}".format(safe_str(x)) for x in sequence])
    elif isinstance(sequence, str):
        logger.warning("Input sequence provided is already in string format. No operation performed")
        string_transform = sequence
    else:
        logger.warning("Input sequence non-iterable %s. Unable to perform concatenation operation", sequence)
        string_transform = sequence
    return string_transform



"""
Async Python Client for generic Biothings API services
"""

from collections.abc import Iterable
from copy import copy
from pathlib import Path
from typing import Dict, Union, Tuple
import logging
import platform
import warnings


import httpx

from biothings_client.client.settings import (
    COMMON_ALIASES,
    COMMON_KWARGS,
    MYCHEM_ALIASES,
    MYCHEM_KWARGS,
    MYDISEASE_ALIASES,
    MYDISEASE_KWARGS,
    MYGENESET_ALIASES,
    MYGENESET_KWARGS,
    MYGENE_ALIASES,
    MYGENE_KWARGS,
    MYTAXON_ALIASES,
    MYTAXON_KWARGS,
    MYVARIANT_ALIASES,
    MYVARIANT_KWARGS,
)
from biothings_client.__version__ import __version__
from biothings_client._dependencies import _CACHING, _CACHING_NOT_SUPPORTED, _PANDAS
from biothings_client.client.exceptions import CachingNotSupportedError, OptionalDependencyImportError
from biothings_client.mixins.gene import MyGeneClientMixin
from biothings_client.mixins.variant import MyVariantClientMixin
from biothings_client.utils.copy import copy_func
from biothings_client.utils.iteration import concatenate_list, iter_n, list_itemcnt

if _PANDAS:
    import pandas

if _CACHING:
    import hishel
    from biothings_client.cache.storage import AsyncBiothingsClientSqlite3Cache

logger = logging.getLogger("biothings.client")
logger.setLevel(logging.INFO)


# Future work:
# Consider use "verbose" settings to control default logging output level
# by doing this instead of using branching throughout the application,
# the business logic can be more concise and more readable.
class AsyncBiothingClient:
    """
    async http client class for accessing the biothings web services
    """

    def __init__(self, url: str = None):
        if url is None:
            url = self._default_url
        self.url = url
        if self.url[-1] == "/":
            self.url = self.url[:-1]

        self.max_query = self._max_query

        # delay and step attributes are for batch queries.
        self.delay = self._delay  # delay is ignored when requests made from cache.
        self.step = self._step

        self.scroll_size = self._scroll_size

        # raise httpx.HTTPError for status_code > 400
        # > but not for 404 on getvariant
        # > set to False to suppress the exceptions.
        self.raise_for_status = True

        self.default_user_agent = (
            "{package_header}/{client_version} (" "python:{python_version} " "httpx:{httpx_version}" ")"
        ).format(
            **{
                "package_header": self._pkg_user_agent_header,
                "client_version": __version__,
                "python_version": platform.python_version(),
                "httpx_version": httpx.__version__,
            }
        )
        self.http_client = None
        self.cache_storage = None
        self.http_client_setup = False
        self.caching_enabled = False

    async def _build_http_client(self, cache_db: Union[str, Path] = None) -> None:
        """
        Builds the async client instance for usage through the lifetime
        of the biothings_client

        This modifies the state of the BiothingsClient instance
        to set the values for the http_client property

        Inputs:
        :param cache_db: pathlike object to the local sqlite3 cache database file

        Outputs:
        :return: None
        """
        if not self.http_client_setup:
            http_transport = httpx.AsyncHTTPTransport()
            self.http_client = httpx.AsyncClient(transport=http_transport)
            self.http_client_setup = True
            self.http_cache_client_setup = False

    async def _build_cache_http_client(self, cache_db: Union[str, Path] = None) -> None:
        """
        Builds the client instance used for caching biothings requests.
        We rebuild the client whenever we enable to caching to ensure
        that we don't create a database files unless the user explicitly
        wants to leverage request caching

        This modifies the state of the BiothingsClient instance
        to set the values for the http_client property and the cache_storage property

        Inputs:
        :param cache_db: pathlike object to the local sqlite3 cache database file

        Outputs:
        :return: None
        """
        if not self.http_client_setup:
            if cache_db is None:
                cache_db = self._default_cache_file
            cache_db = Path(cache_db).resolve().absolute()

            self.cache_storage = AsyncBiothingsClientSqlite3Cache()
            await self.cache_storage.setup_database_connection(cache_db)
            cache_transport = hishel.AsyncCacheTransport(
                transport=httpx.AsyncHTTPTransport(), storage=self.cache_storage
            )
            cache_controller = hishel.Controller(cacheable_methods=["GET", "POST"])
            self.http_client = hishel.AsyncCacheClient(
                controller=cache_controller, transport=cache_transport, storage=self.cache_storage
            )
            self.http_client_setup = True

    async def __del__(self):
        """
        Destructor for the client to ensure that we close any potential
        connections to the cache database
        """
        try:
            if self.http_client is not None:
                await self.http_client.aclose()
        except Exception as gen_exc:
            logger.exception(gen_exc)
            logger.error("Unable to close the httpx client instance %s", self.http_client)

    def use_http(self):
        """
        Use http instead of https for API calls.
        """
        if self.url:
            self.url = self.url.replace("https://", "http://")

    def use_https(self):
        """
        Use https instead of http for API calls. This is the default.
        """
        if self.url:
            self.url = self.url.replace("http://", "https://")

    @staticmethod
    async def _dataframe(obj, dataframe, df_index=True):
        """
        Converts object to DataFrame (pandas)
        """
        if _PANDAS:
            # if dataframe not in ["by_source", "normal"]:
            if dataframe not in [1, 2]:
                raise ValueError("dataframe must be either 1 (using json_normalize) or 2 (using DataFrame.from_dict")

            if "hits" in obj:
                if dataframe == 1:
                    df = pandas.json_normalize(obj["hits"])
                else:
                    df = pandas.DataFrame.from_dict(obj)
            else:
                if dataframe == 1:
                    df = pandas.json_normalize(obj)
                else:
                    df = pandas.DataFrame.from_dict(obj)
            if df_index:
                df = df.set_index("query")
            return df
        else:
            dataframe_library_error = OptionalDependencyImportError(
                optional_function_access="enable dataframe conversion", optional_group="dataframe", libraries=["pandas"]
            )
            raise dataframe_library_error

    async def _get(
        self, url: str, params: dict = None, none_on_404: bool = False, verbose: bool = True
    ) -> Tuple[bool, httpx.Response]:
        """
        Wrapper around the httpx.get method
        """
        await self._build_http_client()
        if params is None:
            params = {}

        debug = params.pop("debug", False)
        return_raw = params.pop("return_raw", False)
        headers = {"user-agent": self.default_user_agent}
        response = await self.http_client.get(
            url=url, params=params, headers=headers, extensions={"cache_disabled": not self.caching_enabled}
        )

        response_extensions = response.extensions
        from_cache = response_extensions.get("from_cache", False)

        if from_cache:
            logger.debug("Cached response %s from %s", response, url)

        if response.is_success:
            if debug or return_raw:
                get_response = (from_cache, response)
            else:
                get_response = (from_cache, response.json())
        else:
            if none_on_404 and response.status_code == 404:
                get_response = (from_cache, None)
            elif self.raise_for_status:
                response.raise_for_status()  # raise httpx._exceptions.HTTPStatusError
        return get_response

    async def _post(self, url: str, params: dict, verbose: bool = True) -> Tuple[bool, httpx.Response]:
        """
        Wrapper around the httpx.post method
        """
        await self._build_http_client()

        if params is None:
            params = {}
        return_raw = params.pop("return_raw", False)
        headers = {"user-agent": self.default_user_agent}
        response = await self.http_client.post(
            url=url, data=params, headers=headers, extensions={"cache_disabled": not self.caching_enabled}
        )

        response_extensions = response.extensions
        from_cache = response_extensions.get("from_cache", False)

        if from_cache:
            logger.debug("Cached response %s from %s", response, url)

        if response.is_success:
            if return_raw:
                post_response = (from_cache, response)
            else:
                response.read()
                post_response = (from_cache, response.json())
        else:
            if self.raise_for_status:
                response.raise_for_status()
            else:
                post_response = (from_cache, response)
        return post_response

    async def _handle_common_kwargs(self, kwargs):
        # handle these common parameters accept field names as the value
        for kw in ["fields", "always_list", "allow_null"]:
            if kw in kwargs:
                kwargs[kw] = concatenate_list(kwargs[kw], quoted=False)
        return kwargs

    async def _repeated_query(self, query_fn, query_li, verbose=True, **fn_kwargs):
        """
        Run query_fn for input query_li in a batch (self.step).
        return a generator of query_result in each batch.
        input query_li can be a list/tuple/iterable
        """
        step = min(self.step, self.max_query)
        i = 0
        for batch, cnt in iter_n(query_li, step, with_cnt=True):
            if verbose:
                logger.info("querying {0}-{1}...".format(i + 1, cnt))
            i = cnt
            _, query_result = await query_fn(batch, **fn_kwargs)
            yield query_result

    async def _metadata(self, verbose=True, **kwargs):
        """
        Return a dictionary of Biothing metadata.
        """
        _url = self.url + self._metadata_endpoint
        _, ret = await self._get(_url, params=kwargs, verbose=verbose)
        return ret

    async def _set_caching(self, cache_db: Union[str, Path] = None, **kwargs) -> None:
        """
        Enable the client caching and creates a local cache database
        for all future requests

        If caching is already enabled then we no-opt

        Inputs:
        :param cache_db: pathlike object to the local sqlite3 cache database file

        Outputs:
        :return: None
        """
        if _CACHING_NOT_SUPPORTED:
            raise CachingNotSupportedError("Caching is only supported for Python 3.8+")

        if _CACHING:
            if not self.caching_enabled:
                try:
                    self.caching_enabled = True
                    self.http_client_setup = False
                    await self._build_cache_http_client()
                    logger.debug("Reset the HTTP client to leverage caching %s", self.http_client)
                    logger.info(
                        (
                            "Enabled client caching: %s\n" 'Future queries will be cached in "%s"',
                            self,
                            self.cache_storage.cache_filepath,
                        )
                    )
                except Exception as gen_exc:
                    logger.exception(gen_exc)
                    logger.error("Unable to enable caching")
                    raise gen_exc
            else:
                logger.warning("Caching already enabled. Skipping for now ...")
        else:
            caching_library_error = OptionalDependencyImportError(
                optional_function_access="enable biothings-client caching",
                optional_group="caching",
                libraries=["anysqlite", "hishel"],
            )
            raise caching_library_error

    async def _stop_caching(self) -> None:
        """
        Disable client caching. The local cache database will be maintained,
        but we will disable cache access when sending requests

        If caching is already disabled then we no-opt

        Inputs:
        :param None

        Outputs:
        :return: None
        """
        if _CACHING_NOT_SUPPORTED:
            raise CachingNotSupportedError("Caching is only supported for Python 3.8+")

        if _CACHING:
            if self.caching_enabled:
                try:
                    await self.cache_storage.clear_cache()
                except Exception as gen_exc:
                    logger.exception(gen_exc)
                    logger.error("Error attempting to clear the local cache database")
                    raise gen_exc

                self.caching_enabled = False
                self.http_client_setup = False
                self._build_http_client()
                logger.debug("Reset the HTTP client to disable caching %s", self.http_client)
                logger.info("Disabled client caching: %s", self)
            else:
                logger.warning("Caching already disabled. Skipping for now ...")
        else:
            caching_library_error = OptionalDependencyImportError(
                optional_function_access="disable biothings-client caching",
                optional_group="caching",
                libraries=["anysqlite", "hishel"],
            )
            raise caching_library_error

    async def _clear_cache(self) -> None:
        """
        Clear the globally installed cache. Caching will stil be enabled,
        but the data stored in the cache stored will be dropped
        """
        if _CACHING_NOT_SUPPORTED:
            raise CachingNotSupportedError("Caching is only supported for Python 3.8+")

        if _CACHING:
            if self.caching_enabled:
                try:
                    await self.cache_storage.clear_cache()
                except Exception as gen_exc:
                    logger.exception(gen_exc)
                    logger.error("Error attempting to clear the local cache database")
                    raise gen_exc
            else:
                logger.warning("Caching already disabled. No local cache database to clear. Skipping for now ...")
        else:
            caching_library_error = OptionalDependencyImportError(
                optional_function_access="clear biothings-client cache",
                optional_group="caching",
                libraries=["anysqlite", "hishel"],
            )
            raise caching_library_error

    async def _get_fields(self, search_term=None, verbose=True):
        """
        Wrapper for /metadata/fields

        **search_term** is a case insensitive string to search for in available field names.
        If not provided, all available fields will be returned.

        .. Hint:: This is useful to find out the field names you need to pass to **fields** parameter of other methods.

        """
        _url = self.url + self._metadata_fields_endpoint
        if search_term:
            params = {"search": search_term}
        else:
            params = {}
        _, ret = await self._get(_url, params=params, verbose=verbose)
        for k, v in ret.items():
            del k
            # Get rid of the notes column information
            if "notes" in v:
                del v["notes"]
        return ret

    async def _getannotation(self, _id, fields=None, **kwargs):
        """
        Return the object given id.
        This is a wrapper for GET query of the biothings annotation service.

        :param _id: an entity id.
        :param fields: fields to return, a list or a comma-separated string.
                       If not provided or **fields="all"**, all available fields
                       are returned.

        :return: an entity object as a dictionary, or None if _id is not found.
        """
        verbose = kwargs.pop("verbose", True)
        if fields:
            kwargs["fields"] = fields
        kwargs = await self._handle_common_kwargs(kwargs)
        _url = self.url + self._annotation_endpoint + str(_id)
        _, ret = await self._get(_url, kwargs, none_on_404=True, verbose=verbose)
        return ret

    async def _getannotations_inner(self, ids, verbose=True, **kwargs):
        id_collection = concatenate_list(ids)
        _kwargs = {"ids": id_collection}
        _kwargs.update(kwargs)
        _url = self.url + self._annotation_endpoint
        return await self._post(_url, _kwargs, verbose=verbose)

    async def _annotations_generator(self, query_fn, ids, verbose=True, **kwargs):
        """
        Function to yield a batch of hits one at a time.
        """
        for hits in self._repeated_query(query_fn, ids, verbose=verbose):
            for hit in hits:
                yield hit

    async def _getannotations(self, ids, fields=None, **kwargs):
        """
        Return the list of annotation objects for the given list of ids.
        This is a wrapper for POST query of the biothings annotation service.

        :param ids: a list/tuple/iterable or a string of ids.
        :param fields: fields to return, a list or a comma-separated string.
                       If not provided or **fields="all"**, all available fields
                       are returned.
        :param as_generator: if True, will yield the results in a generator.
        :param as_dataframe: if True or 1 or 2, return object as DataFrame (requires Pandas).
                                  True or 1: using json_normalize
                                  2        : using DataFrame.from_dict
                                  otherwise: return original json
        :param df_index: if True (default), index returned DataFrame by 'query',
                         otherwise, index by number. Only applicable if as_dataframe=True.

        :return: a list of objects or a pandas DataFrame object (when **as_dataframe** is True)

        .. Hint:: A large list of more than 1000 input ids will be sent to the backend
                  web service in batches (1000 at a time), and then the results will be
                  concatenated together. So, from the user-end, it's exactly the same as
                  passing a shorter list. You don't need to worry about saturating our
                  backend servers.

        .. Hint:: If you need to pass a very large list of input ids, you can pass a generator
                  instead of a full list, which is more memory efficient.
        """
        if isinstance(ids, str):
            ids = ids.split(",") if ids else []
        if not (isinstance(ids, (list, tuple, Iterable))):
            raise ValueError('input "ids" must be a list, tuple or iterable.')
        if fields:
            kwargs["fields"] = fields
        kwargs = await self._handle_common_kwargs(kwargs)
        verbose = kwargs.pop("verbose", True)
        dataframe = kwargs.pop("as_dataframe", None)
        df_index = kwargs.pop("df_index", True)
        generator = kwargs.pop("as_generator", False)
        if dataframe in [True, 1]:
            dataframe = 1
        elif dataframe != 2:
            dataframe = None
        return_raw = kwargs.get("return_raw", False)
        if return_raw:
            dataframe = None

        async def query_fn(ids):
            return await self._getannotations_inner(ids, verbose=verbose, **kwargs)

        if generator:
            return self._annotations_generator(query_fn, ids, verbose=verbose, **kwargs)
        out = []
        async for hits in self._repeated_query(query_fn, ids, verbose=verbose):
            if return_raw:
                out.append(hits)  # hits is the raw response text
            else:
                out.extend(hits)
        if return_raw and len(out) == 1:
            out = out[0]
        if dataframe:
            out = await self._dataframe(out, dataframe, df_index=df_index)
        return out

    async def _query(self, q: str, **kwargs):
        """
        Return the query result.
        This is a wrapper for GET query of biothings query service.

        :param q: a query string.
        :param fields: fields to return, a list or a comma-separated string.
                       If not provided or **fields="all"**, all available fields
                       are returned.
        :param size:   the maximum number of results to return (with a cap
                       of 1000 at the moment). Default: 10.
        :param skip:   the number of results to skip. Default: 0.
        :param sort:   Prefix with "-" for descending order, otherwise in ascending order.
                       Default: sort by matching scores in decending order.
        :param as_dataframe: if True or 1 or 2, return object as DataFrame (requires Pandas).
                                  True or 1: using json_normalize
                                  2        : using DataFrame.from_dict
                                  otherwise: return original json
        :param fetch_all: if True, return a generator to all query results (unsorted).  This can provide a very fast
                          return of all hits from a large query.
                          Server requests are done in blocks of 1000 and yielded individually.  Each 1000 block of
                          results must be yielded within 1 minute, otherwise the request will expire at server side.

        :return: a dictionary with returned variant hits or a pandas DataFrame object (when **as_dataframe** is True)
                 or a generator of all hits (when **fetch_all** is True)

        .. Hint:: By default, **query** method returns the first 10 hits if the matched hits are >10.
                  If the total number of hits are less than 1000, you can increase the value for
                  **size** parameter. For a query that returns more than 1000 hits, you can pass
                  "fetch_all=True" to return a `generator <http://www.learnpython.org/en/Generators>`_
                  of all matching hits (internally, those hits are requested from the server in blocks of 1000).
        """
        _url = self.url + self._query_endpoint
        verbose = kwargs.pop("verbose", True)
        kwargs = await self._handle_common_kwargs(kwargs)
        kwargs.update({"q": q})
        fetch_all = kwargs.get("fetch_all")
        if fetch_all in [True, 1]:
            if kwargs.get("as_dataframe", None) in [True, 1]:
                warnings.warn(
                    "Ignored 'as_dataframe' because 'fetch_all' is specified. "
                    "Too many documents to return as a Dataframe."
                )
            return self._fetch_all(url=_url, verbose=verbose, **kwargs)
        dataframe = kwargs.pop("as_dataframe", None)
        if dataframe in [True, 1]:
            dataframe = 1
        elif dataframe != 2:
            dataframe = None
        _, out = await self._get(_url, kwargs, verbose=verbose)
        if dataframe:
            out = await self._dataframe(out, dataframe, df_index=False)
        return out

    async def _fetch_all(self, url: str, verbose: bool = True, **kwargs):
        """
        Function that returns a generator to results. Assumes that 'q' is in kwargs.
        Implicitly disables caching to ensure we actually hit the endpoint rather than
        pulling from local cache
        """
        logger.warning("fetch_all implicitly disables HTTP request caching")
        restore_caching = False
        if self.caching_enabled:
            restore_caching = True
            try:
                await self.stop_caching()
            except OptionalDependencyImportError as optional_import_error:
                logger.exception(optional_import_error)
                logger.debug("No cache to disable for fetch all. Continuing ...")
            except Exception as gen_exc:
                logger.exception(gen_exc)
                logger.error("Unknown error occured while attempting to disable caching")
                raise gen_exc

        _, batch = await self._get(url, params=kwargs, verbose=verbose)
        if verbose:
            logger.info("Fetching {0} {1} . . .".format(batch["total"], self._optionally_plural_object_type))
        for key in ["q", "fetch_all"]:
            kwargs.pop(key)
        while not batch.get("error", "").startswith("No results to return"):
            if "error" in batch:
                logger.error(batch["error"])
                break
            if "_warning" in batch and verbose:
                logger.warning(batch["_warning"])
            for hit in batch["hits"]:
                yield hit
            kwargs.update({"scroll_id": batch["_scroll_id"]})
            _, batch = await self._get(url, params=kwargs, verbose=verbose)

        if restore_caching:
            logger.debug("re-enabling the client HTTP caching")
            try:
                await self.set_caching()
            except OptionalDependencyImportError as optional_import_error:
                logger.exception(optional_import_error)
                logger.debug("No cache to disable for fetch all. Continuing ...")
            except Exception as gen_exc:
                logger.exception(gen_exc)
                logger.error("Unknown error occured while attempting to disable caching")
                raise gen_exc

    async def _querymany_inner(self, qterms, verbose=True, **kwargs):
        query_term_collection = concatenate_list(qterms)
        _kwargs = {"q": query_term_collection}
        _kwargs.update(kwargs)
        _url = self.url + self._query_endpoint
        return await self._post(_url, params=_kwargs, verbose=verbose)

    async def _querymany(self, qterms, scopes=None, **kwargs):
        """
        Return the batch query result.
        This is a wrapper for POST query of "/query" service.

        :param qterms: a list/tuple/iterable of query terms, or a string of comma-separated query terms.
        :param scopes: specify the type (or types) of identifiers passed to **qterms**,
                       either a list or a comma-separated fields to specify type of input qterms.
        :param fields: fields to return, a list or a comma-separated string.
                       If not provided or **fields="all"**, all available fields
                       are returned.
        :param returnall:   if True, return a dict of all related data, including dup. and missing qterms
        :param verbose:     if True (default), print out information about dup and missing qterms
        :param as_dataframe: if True or 1 or 2, return object as DataFrame (requires Pandas).
                                  True or 1: using json_normalize
                                  2        : using DataFrame.from_dict
                                  otherwise: return original json
        :param df_index: if True (default), index returned DataFrame by 'query',
                         otherwise, index by number. Only applicable if as_dataframe=True.
        :return: a list of matching objects or a pandas DataFrame object.

        .. Hint:: Passing a large list of ids (>1000) to :py:meth:`querymany` is perfectly fine.

        .. Hint:: If you need to pass a very large list of input qterms, you can pass a generator
                  instead of a full list, which is more memory efficient.

        """
        if isinstance(qterms, str):
            qterms = qterms.split(",") if qterms else []
        if not (isinstance(qterms, (list, tuple, Iterable))):
            raise ValueError('input "qterms" must be a list, tuple or iterable.')

        if scopes:
            kwargs["scopes"] = concatenate_list(scopes, quoted=False)
        kwargs = await self._handle_common_kwargs(kwargs)
        returnall = kwargs.pop("returnall", False)
        verbose = kwargs.pop("verbose", True)
        dataframe = kwargs.pop("as_dataframe", None)
        if dataframe in [True, 1]:
            dataframe = 1
        elif dataframe != 2:
            dataframe = None
        df_index = kwargs.pop("df_index", True)
        return_raw = kwargs.get("return_raw", False)
        if return_raw:
            dataframe = None

        out = []
        li_missing = []
        li_dup = []
        li_query = []

        async def query_fn(qterms):
            return await self._querymany_inner(qterms, verbose=verbose, **kwargs)

        async for hits in self._repeated_query(query_fn, qterms, verbose=verbose):
            if return_raw:
                out.append(hits)  # hits is the raw response text
            else:
                out.extend(hits)
                for hit in hits:
                    if hit.get("notfound", False):
                        li_missing.append(hit["query"])
                    else:
                        li_query.append(hit["query"])

        if verbose:
            logger.info("Finished.")
        if return_raw:
            if len(out) == 1:
                out = out[0]
            return out

        # check dup hits
        if li_query:
            li_dup = [(query, cnt) for query, cnt in list_itemcnt(li_query) if cnt > 1]
            del li_query

        if dataframe:
            out = await self._dataframe(out, dataframe, df_index=df_index)
            li_dup_df = pandas.DataFrame.from_records(li_dup, columns=["query", "duplicate hits"])
            li_missing_df = pandas.DataFrame(li_missing, columns=["query"])

        if verbose:
            if li_dup:
                logger.warning("{0} input query terms found dup hits:".format(len(li_dup)) + "\t" + str(li_dup)[:100])
            if li_missing:
                logger.warning(
                    "{0} input query terms found no hit:".format(len(li_missing)) + "\t" + str(li_missing)[:100]
                )

        if returnall:
            if dataframe:
                return {"out": out, "dup": li_dup_df, "missing": li_missing_df}
            else:
                return {"out": out, "dup": li_dup, "missing": li_missing}
        else:
            if verbose and (li_dup or li_missing):
                logger.info('Pass "returnall=True" to return complete lists of duplicate or missing query terms.')
            return out


def get_async_client(biothing_type: str = None, instance: bool = True, *args, **kwargs):
    """
    Function to return a new python asynchronous client for a Biothings API service.

    :param biothing_type: the type of biothing client, currently one of: 'gene', 'variant', 'taxon', 'chem', 'disease', 'geneset'
    :param instance: if True, return an instance of the derived client,
                     if False, return the class of the derived client

    All other args/kwargs are passed to the derived client instantiation (if applicable)
    """
    if not biothing_type:
        url = kwargs.get("url", False)
        if not url:
            raise RuntimeError("No biothings_type or url specified.")
        try:
            url += "metadata" if url.endswith("/") else "/metadata"
            res = httpx.get(url)
            dic = res.json()
            biothing_type = dic.get("biothing_type")
            if isinstance(biothing_type, list):
                if len(biothing_type) == 1:
                    # if a list with only one item, use that item
                    biothing_type = biothing_type[0]
                else:
                    raise RuntimeError("Biothing_type in metadata url is not unique.")
            if not isinstance(biothing_type, str):
                raise RuntimeError("Biothing_type in metadata url is not a valid string.")
        except httpx.RequestError as request_error:
            raise RuntimeError("Cannot access metadata url to determine biothing_type.") from request_error
    else:
        biothing_type = biothing_type.lower()
    if biothing_type not in ASYNC_CLIENT_SETTINGS and not kwargs.get("url", False):
        raise TypeError(
            f"No client named '{biothing_type}', currently available clients are: {list(ASYNC_CLIENT_SETTINGS.keys())}"
        )
    _settings = (
        ASYNC_CLIENT_SETTINGS[biothing_type]
        if biothing_type in ASYNC_CLIENT_SETTINGS
        else generate_async_settings(biothing_type, kwargs.get("url"))
    )
    _class = type(
        _settings["class_name"], tuple([_settings["base_class"]] + _settings["mixins"]), _settings["class_kwargs"]
    )
    for src_attr, target_attr in _settings["attr_aliases"].items():
        if getattr(_class, src_attr, False):
            setattr(_class, target_attr, copy_func(getattr(_class, src_attr), name=target_attr))
    for _name, _docstring in _settings["class_kwargs"]["_docstring_obj"].items():
        _func = getattr(_class, _name, None)
        if _func:
            try:
                _func.__doc__ = _docstring
            except AttributeError:
                _func.__func__.__doc__ = _docstring
    _client = _class(*args, **kwargs) if instance else _class
    return _client


# ***********************************************
# *  Client settings
# *
# *  This object contains the client-specific settings necessary to
# *  instantiate a new biothings client.  The currently supported
# *  clients are the keys of this object.
# *
# *  class - the client Class name
# *  class_kwargs - keyword arguments passed to Class on creation
# *  function_aliases - client specific function aliases in Class
# *  ancestors - a list of classes that Class inherits from
# ***********************************************
ASYNC_CLIENT_SETTINGS = {
    "gene": {
        "class_name": "AsyncMyGeneInfo",
        "class_kwargs": MYGENE_KWARGS,
        "attr_aliases": MYGENE_ALIASES,
        "base_class": AsyncBiothingClient,
        "mixins": [MyGeneClientMixin],
    },
    "variant": {
        "class_name": "AsyncMyVariantInfo",
        "class_kwargs": MYVARIANT_KWARGS,
        "attr_aliases": MYVARIANT_ALIASES,
        "base_class": AsyncBiothingClient,
        "mixins": [MyVariantClientMixin],
    },
    "taxon": {
        "class_name": "AsyncMyTaxonInfo",
        "class_kwargs": MYTAXON_KWARGS,
        "attr_aliases": MYTAXON_ALIASES,
        "base_class": AsyncBiothingClient,
        "mixins": [],
    },
    "drug": {
        "class_name": "AsyncMyChemInfo",
        "class_kwargs": MYCHEM_KWARGS,
        "attr_aliases": MYCHEM_ALIASES,
        "base_class": AsyncBiothingClient,
        "mixins": [],
    },
    "chem": {
        "class_name": "AsyncMyChemInfo",
        "class_kwargs": MYCHEM_KWARGS,
        "attr_aliases": MYCHEM_ALIASES,
        "base_class": AsyncBiothingClient,
        "mixins": [],
    },
    "compound": {
        "class_name": "AsyncMyChemInfo",
        "class_kwargs": MYCHEM_KWARGS,
        "attr_aliases": MYCHEM_ALIASES,
        "base_class": AsyncBiothingClient,
        "mixins": [],
    },
    "disease": {
        "class_name": "AsyncMyDiseaseInfo",
        "class_kwargs": MYDISEASE_KWARGS,
        "attr_aliases": MYDISEASE_ALIASES,
        "base_class": AsyncBiothingClient,
        "mixins": [],
    },
    "geneset": {
        "class_name": "AsyncMyGenesetInfo",
        "class_kwargs": MYGENESET_KWARGS,
        "attr_aliases": MYGENESET_ALIASES,
        "base_class": AsyncBiothingClient,
        "mixins": [],
    },
}


def generate_async_settings(biothing_type: str, url: str) -> Dict:
    """
    Tries to generate a settings dictionary for a client that isn't explicitly listed in
    {CLIENT_SETTTINGS, ASYNC_CLIENT_SETTINGS}

    :param biothing_type: The biothing type to target when generating settings
    :param url: The web url specified in the settings via `default_url`

    :return: Returns a dictionary mapping of the client settings
    :rtype: dict
    """

    def _pluralize(s, optional=True):
        _append = "({})" if optional else "{}"
        return s + _append.format("es") if s.endswith("s") else s + _append.format("s")

    _kwargs = copy(COMMON_KWARGS)
    _aliases = copy(COMMON_ALIASES)
    _kwargs.update(
        {
            "_default_url": url,
            "_annotation_endpoint": "/" + biothing_type.lower() + "/",
            "_optionally_plural_object_type": _pluralize(biothing_type.lower()),
            "_default_cache_file": "my" + biothing_type.lower() + "_cache",
        }
    )
    _aliases.update(
        {
            "_getannotation": "get" + biothing_type.lower(),
            "_getannotations": "get" + _pluralize(biothing_type.lower(), optional=False),
        }
    )
    return {
        "class_name": "Async" + "My" + biothing_type.title() + "Info",
        "class_kwargs": _kwargs,
        "mixins": [],
        "attr_aliases": _aliases,
        "base_class": AsyncBiothingClient,
    }



"""
Generic client for Biothings APIs
"""

from biothings_client.client.asynchronous import AsyncBiothingClient, get_async_client
from biothings_client.client.base import BiothingClient, get_client
from biothings_client.__version__ import __version__
from biothings_client._dependencies import _CACHING, _PANDAS

__all__ = [
    "AsyncBiothingClient",
    "BiothingClient",
    "_CACHING",
    "_PANDAS",
    "__version__",
    "get_async_client",
    "get_client",
]



DOCSTRING = {
    "getvariant": """Return the variant object for the give HGVS-based variant id.
        This is a wrapper for GET query of "/variant/<hgvsid>" service.

        :param vid: an HGVS-based variant id. `More about HGVS id <http://docs.myvariant.info/en/latest/doc/data.html#id-field>`_.
        :param fields: fields to return, a list or a comma-separated string.
                       If not provided or **fields="all"**, all available fields
                       are returned. See `here <http://docs.myvariant.info/en/latest/doc/data.html#available-fields>`__
                       for all available fields.
        :param assembly: specify the human genome assembly used in HGVS-based variant id, "hg19" (default) or "hg38".

        :return: a variant object as a dictionary, or None if vid is not found.

        Example:

        >>> mv.getvariant('chr9:g.107620835G>A')
        >>> mv.getvariant('chr9:g.107620835G>A', fields='dbnsfp.genename')
        >>> mv.getvariant('chr9:g.107620835G>A', fields=['dbnsfp.genename', 'cadd.phred'])
        >>> mv.getvariant('chr9:g.107620835G>A', fields='all')
        >>> mv.getvariant('chr1:g.161362951G>A', assembly='hg38')

        .. Hint:: The supported field names passed to **fields** parameter can be found from
                  any full variant object (without **fields**, or **fields="all"**). Note that field name supports dot
                  notation for nested data structure as well, e.g. you can pass "dbnsfp.genename" or
                  "cadd.phred".
        """,
    "getvariants": """Return the list of variant annotation objects for the given list of hgvs-base varaint ids.
        This is a wrapper for POST query of "/variant" service.

        :param ids: a list/tuple/iterable or a string of comma-separated HGVS ids.
                    `More about hgvs id <http://docs.myvariant.info/en/latest/doc/data.html#id-field>`_.
        :param fields: fields to return, a list or a comma-separated string.
                       If not provided or **fields="all"**, all available fields
                       are returned. See `here <http://docs.myvariant.info/en/latest/doc/data.html#available-fields>`__
                       for all available fields.
        :param assembly: specify the human genome assembly used in HGVS-based variant id, "hg19" (default) or "hg38".
        :param as_generator:  if True, will yield the results in a generator.
        :param as_dataframe: if True or 1 or 2, return object as DataFrame (requires Pandas).
                                  True or 1: using json_normalize
                                  2        : using DataFrame.from_dict
                                  otherwise: return original json
        :param df_index: if True (default), index returned DataFrame by 'query',
                         otherwise, index by number. Only applicable if as_dataframe=True.

        :return: a list of variant objects or a pandas DataFrame object (when **as_dataframe** is True)

        :ref: http://docs.myvariant.info/en/latest/doc/variant_annotation_service.html.

        Example:

        >>> vars = ['chr1:g.866422C>T',
        ...         'chr1:g.876664G>A',
        ...         'chr1:g.69635G>C',
        ...         'chr1:g.69869T>A',
        ...         'chr1:g.881918G>A',
        ...         'chr1:g.865625G>A',
        ...         'chr1:g.69892T>C',
        ...         'chr1:g.879381C>T',
        ...         'chr1:g.878330C>G']
        >>> mv.getvariants(vars, fields="cadd.phred")
        >>> mv.getvariants('chr1:g.876664G>A,chr1:g.881918G>A', fields="all")
        >>> mv.getvariants(['chr1:g.876664G>A', 'chr1:g.881918G>A'], as_dataframe=True)
        >>> mv.getvariants(['chr1:g.161362951G>A', 'chr2:g.51032181G>A'], assembly='hg38')

        .. Hint:: A large list of more than 1000 input ids will be sent to the backend
                  web service in batches (1000 at a time), and then the results will be
                  concatenated together. So, from the user-end, it's exactly the same as
                  passing a shorter list. You don't need to worry about saturating our
                  backend servers.

        .. Hint:: If you need to pass a very large list of input ids, you can pass a generator
                  instead of a full list, which is more memory efficient.
        """,
    "query": """Return  the query result.
        This is a wrapper for GET query of "/query?q=<query>" service.

        :param q: a query string, detailed query syntax `here <http://docs.myvariant.info/en/latest/doc/variant_query_service.html#query-syntax>`__.
        :param fields: fields to return, a list or a comma-separated string.
                       If not provided or **fields="all"**, all available fields
                       are returned. See `here <http://docs.myvariant.info/en/latest/doc/data.html#available-fields>`__
                       for all available fields.
        :param assembly: specify the human genome assembly used for the query, "hg19" (default) or "hg38".
        :param size:   the maximum number of results to return (with a cap
                       of 1000 at the moment). Default: 10.
        :param skip:   the number of results to skip. Default: 0.
        :param sort:   Prefix with "-" for descending order, otherwise in ascending order.
                       Default: sort by matching scores in decending order.
        :param as_dataframe: if True or 1 or 2, return object as DataFrame (requires Pandas).
                                  True or 1: using json_normalize
                                  2        : using DataFrame.from_dict
                                  otherwise: return original json
        :param fetch_all: if True, return a generator to all query results (unsorted).  This can provide a very fast
                          return of all hits from a large query.
                          Server requests are done in blocks of 1000 and yielded individually.  Each 1000 block of
                          results must be yielded within 1 minute, otherwise the request will expire at server side.

        :return: a dictionary with returned variant hits or a pandas DataFrame object (when **as_dataframe** is True)
                 or a generator of all hits (when **fetch_all** is True)
        :ref: http://docs.myvariant.info/en/latest/doc/variant_query_service.html.

        Example:

        >>> mv.query('_exists_:dbsnp AND _exists_:cosmic')
        >>> mv.query('dbnsfp.polyphen2.hdiv.score:>0.99 AND chrom:1')
        >>> mv.query('cadd.phred:>50')
        >>> mv.query('dbnsfp.genename:CDK2', size=5)
        >>> mv.query('dbnsfp.genename:CDK2', size=5, assembly='hg38')
        >>> mv.query('dbnsfp.genename:CDK2', fetch_all=True)
        >>> mv.query('chrX:151073054-151383976')

        .. Hint:: By default, **query** method returns the first 10 hits if the matched hits are >10. If the total number
                  of hits are less than 1000, you can increase the value for **size** parameter. For a query returns
                  more than 1000 hits, you can pass "fetch_all=True" to return a `generator <http://www.learnpython.org/en/Generators>`_
                  of all matching hits (internally, those hits are requested from the server-side in blocks of 1000).
        """,
    "querymany": """Return the batch query result.
        This is a wrapper for POST query of "/query" service.

        :param qterms: a list/tuple/iterable of query terms, or a string of comma-separated query terms.
        :param scopes: specify the type (or types) of identifiers passed to **qterms**, either a list or a comma-separated fields to specify type of
                       input qterms, e.g. "dbsnp.rsid", "clinvar.rcv_accession", ["dbsnp.rsid", "cosmic.cosmic_id"].
                       See `here <http://docs.myvariant.info/en/latest/doc/data.html#available-fields>`__ for full list
                       of supported fields.
        :param fields: fields to return, a list or a comma-separated string.
                       If not provided or **fields="all"**, all available fields
                       are returned. See `here <http://docs.myvariant.info/en/latest/doc/data.html#available-fields>`__
                       for all available fields.
        :param assembly: specify the human genome assembly used for the query, "hg19" (default) or "hg38".
        :param returnall:   if True, return a dict of all related data, including dup. and missing qterms
        :param verbose:     if True (default), print out information about dup and missing qterms
        :param as_dataframe: if True or 1 or 2, return object as DataFrame (requires Pandas).
                                  True or 1: using json_normalize
                                  2        : using DataFrame.from_dict
                                  otherwise: return original json
        :param df_index: if True (default), index returned DataFrame by 'query',
                         otherwise, index by number. Only applicable if as_dataframe=True.
        :return: a list of matching variant objects or a pandas DataFrame object.
        :ref: http://docs.myvariant.info/en/latest/doc/variant_query_service.html for available
              fields, extra *kwargs* and more.

        Example:

        >>> mv.querymany(['rs58991260', 'rs2500'], scopes='dbsnp.rsid')
        >>> mv.querymany(['rs58991260', 'rs2500'], scopes='dbsnp.rsid', assembly='hg38')
        >>> mv.querymany(['RCV000083620', 'RCV000083611', 'RCV000083584'], scopes='clinvar.rcv_accession')
        >>> mv.querymany(['COSM1362966', 'COSM990046', 'COSM1392449'], scopes='cosmic.cosmic_id', fields='cosmic')
        >>> mv.querymany(['COSM1362966', 'COSM990046', 'COSM1392449'], scopes='cosmic.cosmic_id',
        ...              fields='cosmic.tumor_site', as_dataframe=True)

        .. Hint:: :py:meth:`querymany` is perfect for query variants based different ids, e.g. rsid, clinvar ids, etc.

        .. Hint:: Just like :py:meth:`getvariants`, passing a large list of ids (>1000) to :py:meth:`querymany` is perfectly fine.

        .. Hint:: If you need to pass a very large list of input qterms, you can pass a generator
                  instead of a full list, which is more memory efficient.

        """,
    "metadata": """Return a dictionary of MyVariant.info metadata.

        :param assembly: return the metadata for either hg19 or hg38 variants, "hg19" (default) or "hg38".

        Example:

        >>> metadata = mv.metadata()
        >>> metadata = mv.metadata(assembly='hg38')

        """,
    "get_fields": """Wrapper for http://myvariant.info/v1/metadata/fields

        :param search_term: a case insensitive string to search for in available field names.
                            If not provided, all available fields will be returned.

        :param assembly: return the metadata for either hg19 or hg38 variants, "hg19" (default) or "hg38".

        Example:

        >>> mv.get_fields()
        >>> mv.get_fields("rsid")
        >>> mv.get_fields("sift")

        .. Hint:: This is useful to find out the field names you need to pass to **fields** parameter of other methods.

        """,
}



"""
Mirror of the chem tests but two main differences:
> asynchronous
> implemented in pytest for asyncio marker
"""

import json
import logging
import types

import pytest


import biothings_client
from biothings_client.client.definitions import AsyncMyChemInfo
from biothings_client.utils.score import descore


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@pytest.mark.asyncio
async def test_metadata(async_chem_client: AsyncMyChemInfo):
    chem_metadata = await async_chem_client.metadata()
    assert "src" in chem_metadata
    assert "stats" in chem_metadata
    assert "total" in chem_metadata["stats"]


@pytest.mark.asyncio
async def test_getchem(async_chem_client: AsyncMyChemInfo):
    c = await async_chem_client.getchem("ZRALSGWEFCBTJO-UHFFFAOYSA-N")
    assert c["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"
    assert c["chebi"]["name"] == "guanidine"


@pytest.mark.asyncio
async def test_getchem_with_fields(async_chem_client: AsyncMyChemInfo):
    c = await async_chem_client.getchem("7AXV542LZ4", fields="chebi.name,chembl.inchi_key,pubchem.cid")
    assert "_id" in c
    assert "chebi" in c
    assert "name" in c["chebi"]
    assert "chembl" in c
    assert "inchi_key" in c["chembl"]
    assert "pubchem" in c
    assert "cid" in c["pubchem"]


@pytest.mark.asyncio
async def test_curie_id_query(async_chem_client: AsyncMyChemInfo):
    """
    Tests the annotation endpoint support for the biolink CURIE ID.

    If support is enabled then we should retrieve the exact same document for all the provided
    queries
    """
    curie_id_testing_collection = [
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "CHEMBL297569",
            "CHEMBL.COMPOUND:CHEMBL297569",
            "chembl.compound:CHEMBL297569",
            "cHEmbl.ComPOUND:CHEMBL297569",
            "chembl.molecule_chembl_id:CHEMBL297569",
        ),
        (
            "AKUPVPKIFATOBM-UHFFFAOYSA-N",
            "120933777",
            120933777,
            "PUBCHEM.COMPOUND:120933777",
            "pubchem.compound:120933777",
            "PuBcHEm.COMPound:120933777",
            "pubchem.cid:120933777",
        ),
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "CHEBI:CHEBI:57966",
            "chebi:CHEBI:57966",
            "CheBi:CHEBI:57966",
            "chebi.id:CHEBI:57966",
        ),
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "11P2JDE17B",
            "UNII:11P2JDE17B",
            "unii:11P2JDE17B",
            "uNIi:11P2JDE17B",
            "unii.unii:11P2JDE17B",
        ),
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "dB03107",
            "DRUGBANK:dB03107",
            "drugbank:dB03107",
            "DrugBaNK:dB03107",
            "drugbank.id:dB03107",
        ),
    ]

    aggregation_query_groups = []
    for query_collection in curie_id_testing_collection:
        query_result_storage = []
        for similar_query in query_collection:
            query_result = await async_chem_client.getchem(_id=similar_query)
            query_result_storage.append(query_result)

        results_aggregation = [query == query_result_storage[0] for query in query_result_storage[1:]]

        query_result_mapping = dict(zip(query_collection[1:], results_aggregation))
        logger.debug(
            "Comparison to first term %s ->\n%s", query_collection[0], json.dumps(query_result_mapping, indent=4)
        )

        if all(results_aggregation):
            logger.info("Query group %s succeeded", query_collection)
        else:
            logger.error("Query group %s failed", query_collection)

        aggregation_query_groups.append(all(results_aggregation))
    assert all(aggregation_query_groups)


@pytest.mark.asyncio
async def test_multiple_curie_id_query(async_chem_client: AsyncMyChemInfo):
    """
    Tests the annotations endpoint support for the biolink CURIE ID.

    Batch query testing against the POST endpoint to verify that the CURIE ID can work with
    multiple

    If support is enabled then we should retrieve the exact same document for all the provided
    queries
    """
    curie_id_testing_collection = [
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "CHEMBL297569",
            "CHEMBL.COMPOUND:CHEMBL297569",
            "chembl.compound:CHEMBL297569",
            "cHEmbl.ComPOUND:CHEMBL297569",
            "chembl.molecule_chembl_id:CHEMBL297569",
        ),
        (
            "AKUPVPKIFATOBM-UHFFFAOYSA-N",
            "120933777",
            120933777,
            "PUBCHEM.COMPOUND:120933777",
            "pubchem.compound:120933777",
            "PuBcHEm.COMPound:120933777",
            "pubchem.cid:120933777",
        ),
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "CHEBI:CHEBI:57966",
            "chebi:CHEBI:57966",
            "CheBi:CHEBI:57966",
            "chebi.id:CHEBI:57966",
        ),
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "11P2JDE17B",
            "UNII:11P2JDE17B",
            "unii:11P2JDE17B",
            "uNIi:11P2JDE17B",
            "unii.unii:11P2JDE17B",
        ),
        (
            "UCMIRNVEIXFBKS-UHFFFAOYSA-N",
            "dB03107",
            "DRUGBANK:dB03107",
            "drugbank:dB03107",
            "DrugBaNK:dB03107",
            "drugbank.id:dB03107",
        ),
    ]

    results_aggregation = []
    for query_collection in curie_id_testing_collection:
        base_result = await async_chem_client.getchem(_id=query_collection[0])
        query_results = await async_chem_client.getchems(ids=query_collection)
        assert len(query_results) == len(query_collection)

        batch_result = []
        for query_result, query_entry in zip(query_results, query_collection):
            return_query_field = query_result.pop("query")
            assert return_query_field == str(query_entry)
            batch_result.append(base_result == query_result)

        aggregate_result = all(results_aggregation)

        query_result_mapping = dict(zip(query_collection[1:], results_aggregation))
        logger.debug(
            "Comparison to first term %s ->\n%s", query_collection[0], json.dumps(query_result_mapping, indent=4)
        )

        if aggregate_result:
            logger.info("Query group %s succeeded", query_collection)
        else:
            logger.error("Query group %s failed", query_collection)

        results_aggregation.append(aggregate_result)
    assert all(results_aggregation)


@pytest.mark.asyncio
@pytest.mark.xfail
async def get_getdrug(async_chem_client: AsyncMyChemInfo):
    c = await async_chem_client.getdrug("CHEMBL1308")
    assert c["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"
    c = await async_chem_client.getdrug("7AXV542LZ4")
    assert c["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"
    c = await async_chem_client.getdrug("CHEBI:6431")
    assert c["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"

    # PubChem CID
    # not working yet
    c = await async_chem_client.getdrug("CID:1990")
    assert c["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"
    c = await async_chem_client.getdrug("1990")
    assert c["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


@pytest.mark.asyncio
async def test_getchems(async_chem_client: AsyncMyChemInfo):
    c_li = await async_chem_client.getchems(
        ["KTUFNOKKBVMGRW-UHFFFAOYSA-N", "HXHWSAZORRCQMX-UHFFFAOYSA-N", "DQMZLTXERSFNPB-UHFFFAOYSA-N"]
    )
    assert len(c_li) == 3
    assert c_li[0]["_id"] == "KTUFNOKKBVMGRW-UHFFFAOYSA-N"
    assert c_li[1]["_id"] == "HXHWSAZORRCQMX-UHFFFAOYSA-N"
    assert c_li[2]["_id"] == "DQMZLTXERSFNPB-UHFFFAOYSA-N"


@pytest.mark.asyncio
async def test_query(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.query("chebi.name:albendazole", size=5)
    assert "hits" in qres
    assert len(qres["hits"]) == 5


@pytest.mark.asyncio
@pytest.mark.skip(reason="drugbank datasource removed")
async def test_query_drugbank(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.query("drugbank.id:DB00536")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


@pytest.mark.asyncio
async def test_query_chebi(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.query(r"chebi.id:CHEBI\:42820")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


@pytest.mark.asyncio
async def test_query_chembl(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.query('chembl.smiles:"CC(=O)NO"')
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "RRUDCFGSUDOHDG-UHFFFAOYSA-N"


@pytest.mark.asyncio
@pytest.mark.xfail
async def test_query_drugcentral(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.query(
        "drugcentral.drug_use.contraindication.umls_cui:C0023530", fields="drugcentral", size=50
    )
    assert "hits" in qres
    assert len(qres["hits"]) == 50

    # not working yet
    qres = await async_chem_client.query("drugcentral.xrefs.kegg_drug:D00220")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


@pytest.mark.asyncio
async def test_query_pubchem(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.query("pubchem.molecular_formula:C2H5NO2", fields="pubchem", size=20)
    assert "hits" in qres
    assert len(qres["hits"]) == 20

    qres = await async_chem_client.query('pubchem.inchi:"InChI=1S/CH5N3/c2-1(3)4/h(H5,2,3,4)"')
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


@pytest.mark.asyncio
async def test_query_ginas(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.query("ginas.approvalID:JU58VJ6Y3B")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


@pytest.mark.asyncio
async def test_query_pharmgkb(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.query("pharmgkb.id:PA164781028")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


@pytest.mark.asyncio
async def test_query_ndc(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.query('ndc.productndc:"27437-051"')
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "KPQZUUQMTUIKBP-UHFFFAOYSA-N"


@pytest.mark.asyncio
@pytest.mark.xfail
async def test_query_sider(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.query("sider.meddra.umls_id:C0232487", fields="sider", size=5)
    assert "hits" in qres
    assert len(qres["hits"]) == 5
    # Temp disable this check till we fix the data issue
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


@pytest.mark.asyncio
async def test_query_unii(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.query("unii.unii:JU58VJ6Y3B")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


@pytest.mark.asyncio
async def test_query_aeolus(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.query("aeolus.rxcui:50675")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "ZRALSGWEFCBTJO-UHFFFAOYSA-N"


@pytest.mark.asyncio
async def test_query_fetch_all(async_chem_client: AsyncMyChemInfo):
    # fetch_all won't work when caching is used.
    # async_chem_client.stop_caching()
    q = "drugcentral.drug_use.contraindication.umls_cui:C0023530"
    qres = await async_chem_client.query(q, size=0)
    total = qres["total"]

    qres_generator = await async_chem_client.query(q, fields="drugcentral.drug_use", fetch_all=True)
    assert isinstance(qres_generator, types.AsyncGeneratorType)

    async_count = 0
    async for async_res in qres_generator:
        async_count += 1
    assert total == async_count


@pytest.mark.asyncio
async def test_querymany(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.querymany(
        ["ZRALSGWEFCBTJO-UHFFFAOYSA-N", "RRUDCFGSUDOHDG-UHFFFAOYSA-N"], verbose=False
    )
    assert len(qres) == 2

    qres = await async_chem_client.querymany("ZRALSGWEFCBTJO-UHFFFAOYSA-N,RRUDCFGSUDOHDG-UHFFFAOYSA-N", verbose=False)
    assert len(qres) == 2


@pytest.mark.asyncio
async def test_querymany_with_scopes(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.querymany(["CHEBI:31690", "CHEBI:15365"], scopes="chebi.id", verbose=False)
    assert len(qres) == 2

    qres = await async_chem_client.querymany(
        ["CHEMBL374515", "4RZ82L2GY5"], scopes="chembl.molecule_chembl_id,unii.unii", verbose=False
    )
    assert len(qres) >= 2


@pytest.mark.asyncio
async def test_querymany_fields(async_chem_client: AsyncMyChemInfo):
    qres1 = await async_chem_client.querymany(
        ["CHEBI:31690", "CHEBI:15365"],
        scopes="chebi.id",
        fields=["chebi.name", "unii.registry_number"],
        verbose=False,
    )
    assert len(qres1) == 2

    qres2 = await async_chem_client.querymany(
        ["CHEBI:31690", "CHEBI:15365"], scopes="chebi.id", fields="chebi.name,unii.registry_number", verbose=False
    )
    assert len(qres2) == 2

    assert descore(qres1) == descore(qres2)


@pytest.mark.asyncio
async def test_querymany_notfound(async_chem_client: AsyncMyChemInfo):
    qres = await async_chem_client.querymany(["CHEBI:31690", "CHEBI:15365", "NA_TEST"], scopes="chebi.id")
    assert len(qres) == 3
    assert qres[2] == {"query": "NA_TEST", "notfound": True}


@pytest.mark.asyncio
@pytest.mark.skipif(not biothings_client._PANDAS, reason="requires the pandas library")
async def test_querymany_dataframe(async_chem_client: AsyncMyChemInfo):
    from pandas import DataFrame

    query_list1 = [
        "QCYGXOCMWHSXSU-UHFFFAOYSA-N",
        "ADFOMBKCPIMCOO-BTVCFUMJSA-N",
        "DNUTZBZXLPWRJG-UHFFFAOYSA-N",
        "DROLRDZYPMOKLM-BIVLZKPYSA-N",
        "KPBZROQVTHLCDU-GOSISDBHSA-N",
        "UTUUIUQHGDRVPU-UHFFFAOYSA-K",
        "WZWDUEKBAIXVCC-IGHBBLSQSA-N",
        "IAJIIJBMBCZPSW-BDAKNGLRSA-N",
        "NASIOHFAYPRIAC-JTQLQIEISA-N",
        "VGWIQFDQAFSSKA-UHFFFAOYSA-N",
    ]

    qres = await async_chem_client.querymany(query_list1, scopes="_id", fields="pubchem", as_dataframe=True)
    assert isinstance(qres, DataFrame)
    assert "pubchem.inchi" in qres.columns
    assert set(query_list1) == set(qres.index)


@pytest.mark.asyncio
async def test_querymany_step(async_chem_client: AsyncMyChemInfo):
    query_list1 = [
        "QCYGXOCMWHSXSU-UHFFFAOYSA-N",
        "ADFOMBKCPIMCOO-BTVCFUMJSA-N",
        "DNUTZBZXLPWRJG-UHFFFAOYSA-N",
        "DROLRDZYPMOKLM-BIVLZKPYSA-N",
        "KPBZROQVTHLCDU-GOSISDBHSA-N",
        "UTUUIUQHGDRVPU-UHFFFAOYSA-K",
        "WZWDUEKBAIXVCC-IGHBBLSQSA-N",
        "IAJIIJBMBCZPSW-BDAKNGLRSA-N",
        "NASIOHFAYPRIAC-JTQLQIEISA-N",
        "VGWIQFDQAFSSKA-UHFFFAOYSA-N",
    ]

    qres1 = await async_chem_client.querymany(query_list1, scopes="_id", fields="pubchem")
    default_step = async_chem_client.step
    async_chem_client.step = 3
    qres2 = await async_chem_client.querymany(query_list1, scopes="_id", fields="pubchem")
    async_chem_client.step = default_step
    qres1 = descore(sorted(qres1, key=lambda doc: doc["_id"]))
    qres2 = descore(sorted(qres2, key=lambda doc: doc["_id"]))
    assert qres1 == qres2


@pytest.mark.asyncio
async def test_get_fields(async_chem_client: AsyncMyChemInfo):
    fields = await async_chem_client.get_fields()
    assert "chembl.inchi_key" in fields.keys()
    assert "pharmgkb.trade_names" in fields.keys()

    fields = await async_chem_client.get_fields("unii")
    assert "unii.molecular_formula" in fields.keys()



def descore(hit):
    """
    Pops the _score from a hit or hit list - _score can vary slightly between runs causing
    tests to fail.
    """
    if isinstance(hit, list):
        res = []
        for o in hit:
            o.pop("_score", None)
            res.append(o)
        return res
    else:
        hit.pop("_score", None)
        return hit



"""
Custom cache storage handling for the biothings-client
"""

from pathlib import Path
from typing import Optional, Union
import logging
import sqlite3

from biothings_client._dependencies import _CACHING

if _CACHING:
    import anysqlite
    import hishel


logger = logging.getLogger("biothings.client")
logger.setLevel(logging.INFO)


class BiothingsClientSqlite3Cache(hishel.SQLiteStorage):
    """
    Overriden sqlite3 client for some extra functionality

    We have two main properties that we want from this overridden
    class:

    1) The ability to get the cache location. This is accessed via
    the `cache_filepath` property
    2) The ability to clear the cache. This can be performed via the
    `clear_cache` methodcall
    """

    def __init__(
        self,
        serializer: Optional[hishel.BaseSerializer] = None,
        connection: Optional[anysqlite.Connection] = None,
        ttl: Optional[Union[int, float]] = None,
    ) -> None:
        self._cache_filepath: Path = None
        self._connection: Optional[sqlite3.Connection] = connection or None
        self._setup_completed: bool = False
        super().__init__(serializer, connection, ttl)

    def setup_database_connection(self, cache_filepath: Union[str, Path] = None) -> None:
        """
        Establishes the sqlite3 database connection if it hasn't been
        created yet

        Override of the _setup method so that we can specify the database
        file path. Exposed publically so that the user can specify this as well
        along with during biothings_client testing
        """
        if not self._setup_completed:
            if cache_filepath is None:
                home_directory = Path.home()
                cache_directory = home_directory.joinpath(".cache")
                cache_directory.mkdir(parents=True, exist_ok=True)
                cache_filepath.joinpath(".hishel.sqlite")
            self._cache_filepath = cache_filepath.resolve().absolute()

            with self._setup_lock:
                if not self._connection:  # pragma: no cover
                    self._connection = sqlite3.connect(self._cache_filepath, check_same_thread=False)
                table_creation_commnd = "CREATE TABLE IF NOT EXISTS cache(key TEXT, data BLOB, date_created REAL)"
                self._connection.execute(table_creation_commnd)
                self._connection.commit()
                self._setup_completed = True

    def clear_cache(self) -> None:
        """
        Clears the sqlite3 cache

        1) Performs a DELETE to remove all rows without
        dropping the table
        2) Update the auto-increment counter
        3) Perform vacuum operation
        """
        cache_table_name = "cache"
        with self._setup_lock:
            try:
                drop_table_command = f"DELETE FROM {cache_table_name}"
                self._connection.execute(drop_table_command)
                self._connection.commit()
            except sqlite3.OperationalError as operational_error:
                logger.exception(operational_error)
                exception_message = operational_error.args[0]
                missing_cache_table_message = f"no such table: {cache_table_name}"
                if exception_message == missing_cache_table_message:
                    logger.debug("No table [%s] to clear. Skipping ...", cache_table_name)
                else:
                    raise operational_error
            except Exception as gen_exc:
                logger.exception(gen_exc)
                raise gen_exc

            autoincrement_table_name = "SQLITE_SEQUENCE"
            try:
                reset_autoincrement_command = (
                    f"UPDATE {autoincrement_table_name} SET seq = 0 WHERE name = '{cache_table_name}'"
                )
                self._connection.execute(reset_autoincrement_command)
                self._connection.commit()
            except sqlite3.OperationalError as operational_error:
                logger.exception(operational_error)
                exception_message = operational_error.args[0]
                missing_autoincrement_table_message = f"no such table: {autoincrement_table_name}"
                if exception_message == missing_autoincrement_table_message:
                    logger.debug("No table [%s] to update. Skipping ...", autoincrement_table_name)
                else:
                    raise operational_error
            except Exception as gen_exc:
                logger.exception(gen_exc)
                raise gen_exc

            try:
                vacuum_command = "VACUUM"
                self._connection.execute(vacuum_command)
                self._connection.commit()
            except sqlite3.OperationalError as operational_error:
                logger.exception(operational_error)
                raise operational_error
            except Exception as gen_exc:
                logger.exception(gen_exc)
                raise gen_exc

    @property
    def cache_filepath(self) -> Path:
        """
        Returns the filepath for the sqlite3 cache database

        We have either stored it because we generated it ourselves
        via `BiothingsClientSqlite3Storage.database_connection` or we
        have to look it up in the database via the following PRAGMA:
        https://www.sqlite.org/pragma.html#pragma_database_list
        """
        self.setup_database_connection()
        if self._cache_filepath is None:
            pragma_command = "PRAGMA database_list"
            for _, name, filename in self._connection.execute(pragma_command):
                if name == "main" and filename is not None:
                    self._cache_filepath = Path(filename).resolve().absolute()
                    break
        return self._cache_filepath


class AsyncBiothingsClientSqlite3Cache(hishel.AsyncSQLiteStorage):
    """
    Overriden sqlite3 client for some extra functionality

    We have two main properties that we want from this overridden
    class:

    1) The ability to get the cache location. This is accessed via
    the `cache_filepath` property
    2) The ability to clear the cache. This can be performed via the
    `clear_cache` methodcall
    """

    def __init__(
        self,
        serializer: Optional[hishel.BaseSerializer] = None,
        connection: Optional[anysqlite.Connection] = None,
        ttl: Optional[Union[int, float]] = None,
    ) -> None:
        self._cache_filepath = None
        super().__init__(serializer, connection, ttl)

    async def setup_database_connection(self, cache_filepath: Union[str, Path] = None) -> None:
        """
        Establishes the sqlite3 database connection if it hasn't been
        created yet

        Override of the _setup method so that we can specify the database
        file path. Exposed publically so that the user can specify this as well
        along with during biothings_client testing
        """
        if not self._setup_completed:
            if cache_filepath is None:
                home_directory = Path.home()
                cache_directory = home_directory.joinpath(".cache")
                cache_directory.mkdir(parents=True, exist_ok=True)
                cache_filepath.joinpath(".hishel.sqlite")
            self._cache_filepath = cache_filepath.resolve().absolute()

            async with self._setup_lock:
                if not self._connection:  # pragma: no cover
                    self._connection = await anysqlite.connect(self._cache_filepath, check_same_thread=False)
                table_creation_commnd = "CREATE TABLE IF NOT EXISTS cache(key TEXT, data BLOB, date_created REAL)"
                await self._connection.execute(table_creation_commnd)
                await self._connection.commit()
                self._setup_completed = True

    async def clear_cache(self) -> None:
        """
        Clears the sqlite3 cache

        1) Performs a DELETE to remove all rows without
        dropping the table
        """
        async with self._setup_lock:
            cache_table_name = "cache"
            try:
                drop_table_command = f"DELETE FROM {cache_table_name}"
                await self._connection.execute(drop_table_command)
                await self._connection.commit()
            except anysqlite.OperationalError as operational_error:
                logger.exception(operational_error)
                exception_message = operational_error.args[0]
                missing_cache_table_message = f"no such table: {cache_table_name}"
                if exception_message == missing_cache_table_message:
                    logger.debug("No table [%s] to clear. Skipping ...", cache_table_name)
                else:
                    raise operational_error
            except Exception as gen_exc:
                logger.exception(gen_exc)
                raise gen_exc

            autoincrement_table_name = "SQLITE_SEQUENCE"
            try:
                reset_autoincrement_command = (
                    f"UPDATE {autoincrement_table_name} SET seq = 0 WHERE name = '{cache_table_name}'"
                )
                await self._connection.execute(reset_autoincrement_command)
                await self._connection.commit()
            except sqlite3.OperationalError as operational_error:
                logger.exception(operational_error)
                exception_message = operational_error.args[0]
                missing_autoincrement_table_message = f"no such table: {autoincrement_table_name}"
                if exception_message == missing_autoincrement_table_message:
                    logger.debug("No table [%s] to update. Skipping ...", autoincrement_table_name)
                else:
                    raise operational_error
            except Exception as gen_exc:
                logger.exception(gen_exc)
                raise gen_exc

            try:
                vacuum_command = "VACUUM"
                await self._connection.execute(vacuum_command)
                await self._connection.commit()
            except sqlite3.OperationalError as operational_error:
                logger.exception(operational_error)
                raise operational_error
            except Exception as gen_exc:
                logger.exception(gen_exc)
                raise gen_exc

    @property
    async def cache_filepath(self) -> Path:
        """
        Returns the filepath for the sqlite3 cache database

        We have either stored it because we generated it ourselves
        via `BiothingsClientSqlite3Storage.database_connection` or we
        have to look it up in the database via the following PRAGMA:
        https://www.sqlite.org/pragma.html#pragma_database_list
        """
        if self._cache_filepath is None:
            pragma_command = "PRAGMA database_list"
            async for _, name, filename in self._connection.execute(pragma_command):
                if name == "main" and filename is not None:
                    self._cache_filepath = Path(filename).resolve().absolute()
                    break
        return self._cache_filepath



"""
Settings and configuration values for the different biothings-clients
"""

from copy import copy

from biothings_client.docstring.chem import DOCSTRING as CHEM_DOCSTRING
from biothings_client.docstring.gene import DOCSTRING as GENE_DOCSTRING
from biothings_client.docstring.variant import DOCSTRING as VARIANT_DOCSTRING
from biothings_client.utils.variant import MYVARIANT_TOP_LEVEL_JSONLD_URIS


# ***********************************************
# *  Aliases.
# *
# *  (key, value) pairs where the function named "key" in utils.BiothingClient
# *  is aliased as "value" in the returned client class.
# ***********************************************
# Function aliases common to all clients
COMMON_ALIASES = {
    "_clear_cache": "clear_cache",
    "_get_fields": "get_fields",
    "_metadata": "metadata",
    "_query": "query",
    "_querymany": "querymany",
    "_set_caching": "set_caching",
    "_stop_caching": "stop_caching",
}

# Set project specific aliases
MYGENE_ALIASES = copy(COMMON_ALIASES)
MYGENE_ALIASES.update({"_getannotation": "getgene", "_getannotations": "getgenes"})
MYVARIANT_ALIASES = copy(COMMON_ALIASES)
MYVARIANT_ALIASES.update({"_getannotation": "getvariant", "_getannotations": "getvariants"})
MYCHEM_ALIASES = copy(COMMON_ALIASES)
MYCHEM_ALIASES.update({"_getannotation": "getchem", "_getannotations": "getchems"})
MYCHEM_ALIASES.update({"getchem": "getdrug", "getchems": "getdrugs"})
MYTAXON_ALIASES = copy(COMMON_ALIASES)
MYTAXON_ALIASES.update({"_getannotation": "gettaxon", "_getannotations": "gettaxa"})
MYDISEASE_ALIASES = copy(COMMON_ALIASES)
MYDISEASE_ALIASES.update({"_getannotation": "getdisease", "_getannotations": "getdiseases"})
MYGENESET_ALIASES = copy(COMMON_ALIASES)
MYGENESET_ALIASES.update({"_getannotation": "getgeneset", "_getannotations": "getgenesets"})


# ***********************************************
# *  Kwargs.
# *
# *  These keyword arguments are attached to the returned client class
# *  on class creation.
# ***********************************************
# Object creation kwargs common to all clients
COMMON_KWARGS = {
    "_delay": 1,
    "_docstring_obj": {},
    "_max_query": 1000,
    "_metadata_endpoint": "/metadata",
    "_metadata_fields_endpoint": "/metadata/fields",
    "_pkg_user_agent_header": "biothings_client.py",
    "_query_endpoint": "/query/",
    "_scroll_size": 1000,
    "_step": 1000,
    "_top_level_jsonld_uris": [],
}
# project specific kwargs
MYGENE_KWARGS = copy(COMMON_KWARGS)
MYGENE_KWARGS.update(
    {
        "_annotation_endpoint": "/gene/",
        "_default_cache_file": "mygene_cache",
        "_default_url": "https://mygene.info/v3",
        "_docstring_obj": GENE_DOCSTRING,
        "_entity": "gene",
        "_optionally_plural_object_type": "gene(s)",
    }
)
MYVARIANT_KWARGS = copy(COMMON_KWARGS)
MYVARIANT_KWARGS.update(
    {
        "_annotation_endpoint": "/variant/",
        "_default_cache_file": "myvariant_cache",
        "_default_url": "https://myvariant.info/v1",
        "_docstring_obj": VARIANT_DOCSTRING,
        "_entity": "variant",
        "_optionally_plural_object_type": "variant(s)",
        "_top_level_jsonld_uris": MYVARIANT_TOP_LEVEL_JSONLD_URIS,
    }
)
MYCHEM_KWARGS = copy(COMMON_KWARGS)
MYCHEM_KWARGS.update(
    {
        "_annotation_endpoint": "/chem/",
        "_default_cache_file": "mychem_cache",
        "_default_url": "https://mychem.info/v1",
        "_docstring_obj": CHEM_DOCSTRING,
        "_entity": "chem",
        "_optionally_plural_object_type": "chem(s)",
    }
)
MYDISEASE_KWARGS = copy(COMMON_KWARGS)
MYDISEASE_KWARGS.update(
    {
        "_annotation_endpoint": "/disease/",
        "_default_cache_file": "mydisease_cache",
        "_default_url": "https://mydisease.info/v1",
        "_entity": "disease",
        "_optionally_plural_object_type": "disease(s)",
    }
)
MYTAXON_KWARGS = copy(COMMON_KWARGS)
MYTAXON_KWARGS.update(
    {
        "_annotation_endpoint": "/taxon/",
        "_default_cache_file": "mytaxon_cache",
        "_default_url": "https://t.biothings.io/v1",
        "_entity": "taxon",
        "_optionally_plural_object_type": "taxon/taxa",
    }
)
MYGENESET_KWARGS = copy(COMMON_KWARGS)
MYGENESET_KWARGS.update(
    {
        "_annotation_endpoint": "/geneset/",
        "_default_cache_file": "mygeneset_cache",
        "_default_url": "https://mygeneset.info/v1",
        "_entity": "geneset",
        "_optionally_plural_object_type": "geneset(s)",
    }
)



from biothings_client.client import get_client, get_async_client


### MYGENE CLIENTS
class MyGeneInfo(get_client("gene", instance=False)):
    pass


class AsyncMyGeneInfo(get_async_client("gene", instance=False)):
    pass


### MYVARIANT CLIENTS
class MyVariantInfo(get_client("variant", instance=False)):
    pass


class AsyncMyVariantInfo(get_async_client("variant", instance=False)):
    pass


### MYCHEM CLIENTS
class MyChemInfo(get_client("chem", instance=False)):
    pass


class AsyncMyChemInfo(get_async_client("chem", instance=False)):
    pass


### MYDISEASE CLIENTS
class MyDiseaseInfo(get_client("disease", instance=False)):
    pass


class AsyncMyDiseaseInfo(get_async_client("disease", instance=False)):
    pass


### MYTAXON CLIENTS
class MyTaxonInfo(get_client("taxon", instance=False)):
    pass


class AsyncMyTaxonInfo(get_async_client("taxon", instance=False)):
    pass


### MYGENESET CLIENTS
class MyGenesetInfo(get_client("geneset", instance=False)):
    pass


class AsyncMyGenesetInfo(get_async_client("geneset", instance=False)):
    pass



from pyld import jsonld


def nquads_transform(doc):
    t = jsonld.JsonLdProcessor()
    nquads = t.parse_nquads(jsonld.to_rdf(doc, {"format": "application/nquads"}))["@default"]

    return nquads


def get_value_and_node(nquads, uri):
    return tuple(
        zip(*[(i["subject"]["value"], i["object"]["value"]) for i in nquads if i["predicate"]["value"] == uri])
    )


def find_top_level_uri(nquads_id, nquads, top_level_uris):
    for item in nquads:
        if item["object"]["value"] == nquads_id:
            if item["predicate"]["value"] in top_level_uris:
                uri = item["predicate"]["value"]
            elif item["predicate"]["value"] not in top_level_uris:
                uri = find_top_level_uri(item["subject"]["value"], nquads, top_level_uris)
            else:  # this can never happen?
                print("couldn't find top level uri")
    return uri


def fetch_value_source(client, _id, uri):
    doc = client._getannotation(_id, jsonld=True)
    nquads = nquads_transform(doc)
    (node, value) = get_value_and_node(nquads, uri)
    source = [find_top_level_uri(item, nquads, client._top_level_jsonld_uris) for item in node]
    result = [i + " " + j for i, j in zip(value, source)]
    return result


def get_uri_list(context):
    uri_path_dict = {}

    for path, v in context.items():
        for field_name, value in v["@context"].items():
            new_path = path.replace("/", ".") + "." + field_name
            if value not in uri_path_dict:
                uri_path_dict[value] = [new_path]
            else:
                uri_path_dict[value].append(new_path)
    return uri_path_dict


def query_by_uri(client, uri, value, context):
    context.pop("root")
    query_string = ""
    path_list = get_uri_list(context)[uri]
    for _item in path_list:
        query_string = query_string + " " + _item + ":" + value + " OR"
    query_string = query_string.strip(" OR")
    return client.query(query_string)



from biothings_client.client.asynchronous import AsyncBiothingClient, get_async_client
from biothings_client.client.base import BiothingClient, get_client


__all__ = ["AsyncBiothingClient", "BiothingClient", "get_async_client", "get_client"]



DOCSTRING = {
    "getgene": """Return the gene object for the give geneid.
        This is a wrapper for GET query of "/gene/<geneid>" service.

        :param geneid: entrez/ensembl gene id, entrez gene id can be either
                       a string or integer
        :param fields: fields to return, a list or a comma-separated string.
                        If **fields="all"**, all available fields are returned
        :param species: optionally, you can pass comma-separated species names
                        or taxonomy ids
        :param email: optionally, pass your email to help us to track usage

        :return: a gene object as a dictionary, or None if geneid is not valid.

        :ref: http://docs.mygene.info/en/latest/doc/data.html for available
             fields, extra *kwargs* and more.

        Example:

        >>> mg.getgene(1017, email='abc@example.com')
        >>> mg.getgene('1017', fields='symbol,name,entrezgene,refseq')
        >>> mg.getgene('1017', fields='symbol,name,entrezgene,refseq.rna')
        >>> mg.getgene('1017', fields=['symbol', 'name', 'pathway.kegg'])
        >>> mg.getgene('ENSG00000123374', fields='all')

        .. Hint:: The supported field names passed to **fields** parameter can be found from
                  any full gene object (when **fields="all"**). Note that field name supports dot
                  notation for nested data structure as well, e.g. you can pass "refseq.rna" or
                  "pathway.kegg".
        """,
    "getgenes": """Return the list of gene objects for the given list of geneids.
        This is a wrapper for POST query of "/gene" service.

        :param geneids: a list/tuple/iterable or comma-separated entrez/ensembl gene ids
        :param fields: fields to return, a list or a comma-separated string.
                        If **fields="all"**, all available fields are returned
        :param species: optionally, you can pass comma-separated species names
                        or taxonomy ids
        :param email: optionally, pass your email to help us to track usage
        :param as_dataframe: if True, return object as DataFrame (requires Pandas).
        :param df_index: if True (default), index returned DataFrame by 'query',
                         otherwise, index by number. Only applicable if as_dataframe=True.

        :return: a list of gene objects or a pandas DataFrame object (when **as_dataframe** is True)

        :ref: http://mygene.info/doc/annotation_service.html for available
                fields, extra *kwargs* and more.

        Example:

        >>> mg.getgenes([1017, '1018','ENSG00000148795'], email='abc@example.com')
        >>> mg.getgenes([1017, '1018','ENSG00000148795'], fields="entrezgene,uniprot")
        >>> mg.getgenes([1017, '1018','ENSG00000148795'], fields="all")
        >>> mg.getgenes([1017, '1018','ENSG00000148795'], as_dataframe=True)

        .. Hint:: A large list of more than 1000 input ids will be sent to the backend
                  web service in batches (1000 at a time), and then the results will be
                  concatenated together. So, from the user-end, it's exactly the same as
                  passing a shorter list. You don't need to worry about saturating our
                  backend servers.
        """,
    "query": """Return  the query result.
        This is a wrapper for GET query of "/query?q=<query>" service.

        :param q: a query string, detailed query syntax `here <http://mygene.info/doc/query_service.html#query-syntax>`_
        :param fields: fields to return, a list or a comma-separated string.
                        If **fields="all"**, all available fields are returned
        :param species: optionally, you can pass comma-separated species names
                        or taxonomy ids. Default: human,mouse,rat.
        :param size:   the maximum number of results to return (with a cap
                       of 1000 at the moment). Default: 10.
        :param skip:   the number of results to skip. Default: 0.
        :param sort:   Prefix with "-" for descending order, otherwise in ascending order.
                       Default: sort by matching scores in decending order.
        :param entrezonly: if True, return only matching entrez genes, otherwise, including matching
                           Ensemble-only genes (those have no matching entrez genes).
        :param email: optionally, pass your email to help us to track usage
        :param as_dataframe: if True, return object as DataFrame (requires Pandas).
        :param df_index: if True (default), index returned DataFrame by 'query',
                         otherwise, index by number. Only applicable if as_dataframe=True.
        :param fetch_all: if True, return a generator to all query results (unsorted).  This can provide a very fast return of
                         all hits from a large query.
                         Server requests are done in blocks of 1000 and yielded individually.  Each 1000 block of results
                         must be yielded within 1 minute, otherwise the request will expire on the server side.

        :return: a dictionary with returned gene hits or a pandas DataFrame object (when **as_dataframe** is True)

        :ref: http://mygene.info/doc/query_service.html for available
              fields, extra *kwargs* and more.

        Example:

        >>> mg.query('cdk2')
        >>> mg.query('reporter:1000_at')
        >>> mg.query('symbol:cdk2', species='human')
        >>> mg.query('symbol:cdk*', species=10090, size=5, as_dataframe=True)
        >>> mg.query('q=chrX:151073054-151383976', species=9606)

        """,
    "querymany": """Return the batch query result.
        This is a wrapper for POST query of "/query" service.

        :param qterms: a list/tuple/iterable of query terms, or a string of comma-separated query terms.
        :param scopes:  type of types of identifiers, either a list or a comma-separated fields to specify type of
                       input qterms, e.g. "entrezgene", "entrezgene,symbol", ["ensemblgene", "symbol"].
                       Refer to `official MyGene.info docs <http://mygene.info/doc/query_service.html#available_fields>`_ for full list
                       of fields.
        :param fields: fields to return, a list or a comma-separated string.
                        If **fields="all"**, all available fields are returned
        :param species: optionally, you can pass comma-separated species names
                          or taxonomy ids. Default: human,mouse,rat.
        :param entrezonly:  if True, return only matching entrez genes, otherwise, including matching
                             Ensemble-only genes (those have no matching entrez genes).

        :param returnall:   if True, return a dict of all related data, including dup. and missing qterms
        :param verbose:     if True (default), print out infomation about dup and missing qterms
        :param email: optionally, pass your email to help us to track usage
        :param as_dataframe: if True, return object as DataFrame (requires Pandas).
        :param df_index: if True (default), index returned DataFrame by 'query',
                         otherwise, index by number. Only applicable if as_dataframe=True.

        :return: a list of gene objects or a pandas DataFrame object (when **as_dataframe** is True)

        :ref: http://mygene.info/doc/query_service.html for available
              fields, extra *kwargs* and more.

        Example:

        >>> mg.querymany(['DDX26B', 'CCDC83'], scopes='symbol', species=9606)
        >>> mg.querymany(['1255_g_at', '1294_at', '1316_at', '1320_at'], scopes='reporter')
        >>> mg.querymany(['NM_003466', 'CDK2', 695, '1320_at', 'Q08345'],
        ...              scopes='refseq,symbol,entrezgene,reporter,uniprot', species='human')
        >>> mg.querymany(['1255_g_at', '1294_at', '1316_at', '1320_at'], scopes='reporter',
        ...              fields='ensembl.gene,symbol', as_dataframe=True)

        .. Hint:: :py:meth:`querymany` is perfect for doing id mappings.

        .. Hint:: Just like :py:meth:`getgenes`, passing a large list of ids (>1000) to :py:meth:`querymany` is perfectly fine.

        """,
    "metadata": """Return a dictionary of MyGene.info metadata.

        Example:

        >>> metadata = mg.metadata

        """,
    "get_fields": """Return all available fields can be return from MyGene.info services.

        This is a wrapper for http://mygene.info/metadata/fields

        :param search_term: an optional string to search (case insensitive) for matching field names.
                            If not provided, all available fields will be returned.

        Example:

        >>> mv.get_fields()
        >>> mv.get_fields("uniprot")
        >>> mv.get_fields("refseq")
        >>> mv.get_fields("kegg")

        .. Hint:: This is useful to find out the field names you need to pass to **fields** parameter of other methods.
        """,
}
